import functools
import uuid
from typing import Any

import fastapi
import sentry_sdk
from fastapi.responses import JSONResponse

from ffun.core import api, errors, logging

logger = logging.get_module_logger()


async def _handle_api_error(request: fastapi.Request, error: errors.APIError) -> JSONResponse:
    # TODO: improve error processing

    api_error = api.APIError(code=error.code, message=error.message)  # type: ignore

    logger.info("api_error", code=error.code, message=error.message)  # type: ignore

    request.state.api_error_code = error.code  # type: ignore

    return JSONResponse(status_code=200, content=api_error.model_dump())


async def _handle_unexpected_error(request: fastapi.Request, error: BaseException) -> JSONResponse:
    # TODO: improve error processing

    code = error.__class__.__name__

    api_error = api.APIError(code=code, message="An unexpected error appeared. We are working on fixing it.")

    request.state.internal_error_code = code

    sentry_sdk.capture_exception(error)

    logger.exception(error.__class__.__name__, sentry_skip=True)

    return JSONResponse(status_code=500, content=api_error.model_dump())


# TODO: move somewhere?
def initialize_error_processors(app: fastapi.FastAPI) -> fastapi.FastAPI:
    app.exception_handler(errors.Error)(_handle_api_error)
    return app


async def final_errors_middleware(request: fastapi.Request, call_next: Any) -> fastapi.Response:
    try:
        return await call_next(request)  # type: ignore
    except errors.APIError as e:
        return await _handle_api_error(request, e)
    except BaseException as e:
        return await _handle_unexpected_error(request, e)


async def request_id_middleware(request: fastapi.Request, call_next: Any) -> fastapi.Response:
    with logging.bound_log_args(request_uid=uuid.uuid4().hex):
        return await call_next(request)  # type: ignore


def _normalize_path(url: str) -> str | None:
    url = url.lower().rstrip("/")

    if "//" in url:
        return None

    return url


@functools.cache
def _existed_route_urls(app: fastapi.FastAPI) -> set[str]:
    # cache should be all right here because app is singleton

    urls = set()

    for route in app.routes:
        url = _normalize_path(route.path)  # type: ignore

        if url is None:
            raise NotImplementedError()

        if "{" in url:
            raise NotImplementedError()

        urls.add(url)

    return urls


async def request_measure_middleware(request: fastapi.Request, call_next: Any) -> fastapi.Response:
    app = request.scope.get("app")

    assert app is not None

    path = _normalize_path(request.scope.get("path"))  # type: ignore

    if path not in _existed_route_urls(app):
        path = "wrong"

    with logger.measure_block_time("request_time", http_path=path) as extra_labels:
        extra_labels["result"] = "success"
        extra_labels["status_code"] = None
        extra_labels["error_code"] = None

        response = await call_next(request)

        extra_labels["status_code"] = response.status_code

        if hasattr(request.state, "api_error_code"):
            extra_labels["error_code"] = request.state.api_error_code
            extra_labels["result"] = "api_error"

        if hasattr(request.state, "internal_error_code"):
            extra_labels["error_code"] = request.state.internal_error_code
            extra_labels["result"] = "internal_error"

        assert isinstance(response, fastapi.Response)

        return response



import uuid

from ffun.core import utils
from ffun.core.postgresql import execute
from ffun.domain.entities import EntryId
from ffun.librarian import operations
from ffun.librarian.entities import ProcessorPointer


async def end_processor_pointer(processor_id: int) -> ProcessorPointer:
    await operations.get_or_create_pointer(processor_id=processor_id)

    next_pointer = ProcessorPointer(
        processor_id=processor_id, pointer_created_at=utils.now(), pointer_entry_id=EntryId(uuid.UUID(int=0))
    )

    await operations.save_pointer(execute, pointer=next_pointer)

    return next_pointer



from typing import Type

import pytest

from ffun.google import errors
from ffun.google.provider_interface import track_key_status
from ffun.llms_framework.entities import KeyStatus
from ffun.llms_framework.keys_statuses import Statuses


class TestTrackKeyStatus:

    def test_works(self, api_key_statuses: Statuses) -> None:
        with track_key_status("key_1", api_key_statuses):
            pass

        assert api_key_statuses.get("key_1") == KeyStatus.works

    @pytest.mark.parametrize("exception", [errors.AuthError])
    def test_authentication_error(self, exception: Type[Exception], api_key_statuses: Statuses) -> None:
        with pytest.raises(exception):
            with track_key_status("key_1", api_key_statuses):
                raise exception()

        assert api_key_statuses.get("key_1") == KeyStatus.broken

    @pytest.mark.parametrize("exception", [errors.QuotaError])
    def test_quota_error(self, exception: Type[Exception], api_key_statuses: Statuses) -> None:
        with pytest.raises(exception):
            with track_key_status("key_1", api_key_statuses):
                raise exception()

        assert api_key_statuses.get("key_1") == KeyStatus.quota

    def test_unknown_client_error(self, api_key_statuses: Statuses) -> None:

        class FakeError(errors.ClientError):
            pass

        with pytest.raises(FakeError):
            with track_key_status("key_1", api_key_statuses):
                raise FakeError()

        assert api_key_statuses.get("key_1") == KeyStatus.unknown

    def test_non_google_error(self, api_key_statuses: Statuses) -> None:
        class FakeError(Exception):
            pass

        with pytest.raises(FakeError):
            with track_key_status("key_1", api_key_statuses):
                raise FakeError()

        assert api_key_statuses.get("key_1") == KeyStatus.unknown



import asyncio
import pprint

from ffun.cli.application import app
from ffun.core.utils import discover_submodules


async def run() -> None:  # noqa: CCR001
    settings = []

    for component in discover_submodules("ffun"):
        for module in discover_submodules(component.__name__):
            if module.__name__.endswith(".settings") and hasattr(module, "settings"):
                settings.append(module)

    for module in settings:
        print("*" * 40)  # noqa: T201
        print(f"* {module.__name__}")  # noqa: T201
        print("*" * 40)  # noqa: T201
        pprint.pprint(module.settings.dict())  # noqa: T203
        print()  # noqa: T201


@app.command()
def print_configs() -> None:
    asyncio.run(run())



"""
excluded-tags-for-rules
"""

from typing import Any, Iterable

from psycopg import Connection
from yoyo import step

__depends__ = {"20230813_01_l7qop-updated-at-field"}


def _key_from_tags(required_tags: Iterable[int], excluded_tags: Iterable[int]) -> str:
    return ",".join(map(str, required_tags)) + "|" + ",".join(map(str, excluded_tags))


def apply_step(conn: Connection[Any]) -> None:
    cursor = conn.cursor()

    cursor.execute("ALTER TABLE s_rules RENAME COLUMN tags TO required_tags")
    cursor.execute("ALTER TABLE s_rules ADD COLUMN excluded_tags BIGINT[] DEFAULT ARRAY[]::BIGINT[]")

    result = cursor.execute("SELECT id, required_tags FROM s_rules").fetchall()

    for row in result:
        cursor.execute(
            "UPDATE s_rules SET key = %(key)s, excluded_tags = ARRAY[]::BIGINT[] WHERE id = %(id)s",
            {"id": row[0], "key": _key_from_tags(row[1], [])},
        )

    cursor.execute("ALTER TABLE s_rules ALTER COLUMN excluded_tags SET NOT NULL")
    cursor.execute("ALTER TABLE s_rules ALTER COLUMN excluded_tags DROP DEFAULT")


def rollback_step(conn: Connection[dict[str, Any]]) -> None:
    cursor = conn.cursor()

    cursor.execute("ALTER TABLE s_rules RENAME COLUMN required_tags TO tags")
    cursor.execute("ALTER TABLE s_rules DROP COLUMN excluded_tags")


steps = [step(apply_step, rollback_step)]



import datetime
import decimal
import enum
from typing import Annotated, NewType

import pydantic

from ffun.core.entities import BaseEntity
from ffun.domain.datetime_intervals import month_interval_start
from ffun.domain.entities import UserId
from ffun.feeds.entities import FeedId

LLMApiKey = NewType("LLMApiKey", str)
LLMUserApiKey = NewType("LLMUserApiKey", LLMApiKey)
LLMCollectionApiKey = NewType("LLMCollectionApiKey", LLMApiKey)
LLMGeneralApiKey = NewType("LLMGeneralApiKey", LLMApiKey)

USDCost = NewType("USDCost", decimal.Decimal)

LLMTokens = NewType("LLMTokens", int)


class LLMProvider(enum.StrEnum):
    test = "test"
    openai = "openai"
    google = "google"


class ModelInfo(pydantic.BaseModel):
    provider: LLMProvider
    name: str
    max_context_size: LLMTokens
    max_return_tokens: LLMTokens

    # protection from overuse/overspend
    max_tokens_per_entry: LLMTokens

    input_1m_tokens_cost: USDCost
    output_1m_tokens_cost: USDCost

    # TODO: test
    def tokens_cost(self, input_tokens: LLMTokens, output_tokens: LLMTokens) -> USDCost:
        cost = (
            self.input_1m_tokens_cost * input_tokens / 1_000_000
            + self.output_1m_tokens_cost * output_tokens / 1_000_000
        )
        return USDCost(cost)


class KeyStatus(str, enum.Enum):
    works = "works"
    broken = "broken"
    quota = "quota"
    unknown = "unknown"


# Generally, different LLM providers can have different config parameters with different names.
# But for simplicity, for now, we will use the unified config for all providers.
# We'll split it later if needed.
class LLMConfiguration(BaseEntity):
    model: str
    system: Annotated[str, pydantic.StringConstraints(strip_whitespace=True)]
    max_return_tokens: LLMTokens
    text_parts_intersection: int
    temperature: float
    top_p: float
    presence_penalty: float
    frequency_penalty: float


class ChatRequest(BaseEntity):
    pass


class ChatResponse(BaseEntity):

    def response_content(self) -> str:
        raise NotImplementedError("Must be implemented in subclasses")

    def input_tokens(self) -> int:
        raise NotImplementedError("Must be implemented in subclasses")

    def output_tokens(self) -> int:
        raise NotImplementedError("Must be implemented in subclasses")


class APIKeyUsage(BaseEntity):
    model_config = pydantic.ConfigDict(frozen=False)

    provider: LLMProvider
    user_id: UserId | None
    api_key: LLMApiKey
    reserved_cost: USDCost
    input_tokens: LLMTokens | None
    output_tokens: LLMTokens | None
    used_cost: USDCost | None
    interval_started_at: datetime.datetime

    def cost_to_register(self) -> USDCost:
        if self.used_cost:
            return self.used_cost

        return self.reserved_cost


class UserKeyInfo(BaseEntity):
    user_id: UserId
    api_key: LLMApiKey | None
    max_tokens_cost_in_month: USDCost
    process_entries_not_older_than: datetime.timedelta
    cost_used: USDCost


class SelectKeyContext(BaseEntity):
    llm_config: LLMConfiguration
    feed_ids: set[FeedId]
    entry_age: datetime.timedelta
    reserved_cost: USDCost
    interval_started_at: datetime.datetime = pydantic.Field(default_factory=month_interval_start)
    collections_api_key: LLMCollectionApiKey | None
    general_api_key: LLMGeneralApiKey | None



import datetime
from typing import Any, Iterable

import psycopg
from pypika import PostgreSQLQuery

from ffun.core import logging, utils
from ffun.core.postgresql import ExecuteType, execute, run_in_transaction
from ffun.domain import urls as domain_urls
from ffun.domain.domain import new_source_id
from ffun.domain.entities import SourceId, UrlUid
from ffun.feeds.entities import Feed, FeedError, FeedId, FeedState

logger = logging.get_module_logger()


def row_to_feed(row: dict[str, Any]) -> Feed:
    return Feed(
        id=row["id"],
        source_id=row["source_id"],
        url=row["url"],
        state=FeedState(row["state"]),
        last_error=FeedError(row["last_error"]) if row["last_error"] else None,
        load_attempted_at=row["load_attempted_at"],
        loaded_at=row["loaded_at"],
        title=row["title"],
        description=row["description"],
    )


async def save_feed(feed: Feed) -> FeedId:
    sql = """
    INSERT INTO f_feeds (id, source_id, url, state, title, description, uid)
    VALUES (%(id)s, %(source_id)s, %(url)s, %(state)s, %(title)s, %(description)s, %(uid)s)
    """

    uid = domain_urls.url_to_uid(feed.url)

    try:
        await execute(
            sql,
            {
                "id": feed.id,
                "source_id": feed.source_id,
                "url": feed.url,
                "state": feed.state,
                "title": feed.title,
                "description": feed.description,
                "uid": uid,
            },
        )

        return feed.id
    except psycopg.errors.UniqueViolation as e:
        logger.warning("unique_violation_while_saving_feed", feed=feed)

        found_ids = await get_feed_ids_by_uids([uid])

        if not found_ids:
            raise NotImplementedError("something went wrong") from e

        return found_ids[uid]


@run_in_transaction
async def get_next_feeds_to_load(execute: ExecuteType, number: int, loaded_before: datetime.datetime) -> list[Feed]:
    sql = """
    SELECT id
    FROM f_feeds
    WHERE load_attempted_at IS NULL OR load_attempted_at <= %(loaded_before)s
    ORDER BY load_attempted_at ASC NULLS FIRST
    LIMIT %(number)s
    FOR UPDATE SKIP LOCKED
    """

    rows = await execute(sql, {"number": number, "loaded_before": loaded_before})

    ids = [row["id"] for row in rows]

    sql = """
    UPDATE f_feeds
    SET load_attempted_at = NOW(),
        updated_at = NOW()
    WHERE id = ANY(%(ids)s)
    RETURNING *
    """

    rows = await execute(sql, {"ids": ids})

    return [row_to_feed(row) for row in rows]


async def update_feed_info(feed_id: FeedId, title: str, description: str) -> None:
    sql = """
    UPDATE f_feeds
    SET title = %(title)s,
        description = %(description)s,
        updated_at = NOW()
    WHERE id = %(id)s
    """

    await execute(sql, {"id": feed_id, "title": title, "description": description})


async def mark_feed_as_loaded(feed_id: FeedId, loaded_at: datetime.datetime | None = None) -> None:
    if loaded_at is None:
        loaded_at = utils.now()

    sql = """
    UPDATE f_feeds
    SET state = %(state)s,
        last_error = NULL,
        loaded_at = %(loaded_at)s,
        updated_at = NOW()
    WHERE id = %(id)s
    """

    await execute(sql, {"id": feed_id, "state": FeedState.loaded, "loaded_at": loaded_at})


async def mark_feed_as_failed(feed_id: FeedId, state: FeedState, error: FeedError) -> None:
    sql = """
    UPDATE f_feeds
    SET state = %(state)s,
        last_error = %(error)s,
        updated_at = NOW()
    WHERE id = %(id)s
    """

    await execute(sql, {"id": feed_id, "state": state, "error": error})


async def mark_feed_as_orphaned(feed_id: FeedId) -> None:
    sql = """
    UPDATE f_feeds
    SET state = %(state)s,
        last_error = NULL,
        updated_at = NOW()
    WHERE id = %(id)s
    """

    await execute(sql, {"id": feed_id, "state": FeedState.orphaned})


async def get_orphaned_feeds(limit: int, loaded_before: datetime.datetime) -> list[FeedId]:
    sql = """
        SELECT *
        FROM f_feeds
        WHERE state = %(state)s AND
              (loaded_at IS NULL OR loaded_at <= %(loaded_before)s)
        ORDER BY created_at ASC
        LIMIT %(limit)s
    """

    rows = await execute(sql, {"state": FeedState.orphaned, "limit": limit, "loaded_before": loaded_before})

    return [row["id"] for row in rows]


async def get_feeds(ids: Iterable[FeedId]) -> list[Feed]:
    sql = """
    SELECT *
    FROM f_feeds
    WHERE id = ANY(%(ids)s)
    """

    rows = await execute(sql, {"ids": list(ids)})

    return [row_to_feed(row) for row in rows]


async def get_feed_ids_by_uids(uids: Iterable[str]) -> dict[UrlUid, FeedId]:
    sql = """
    SELECT id, uid
    FROM f_feeds
    WHERE uid = ANY(%(uids)s)
    """

    rows = await execute(sql, {"uids": list(uids)})

    return {row["uid"]: row["id"] for row in rows}


async def get_source_ids(source_uids: Iterable[str]) -> dict[str, SourceId]:

    if not source_uids:
        return {}

    query = PostgreSQLQuery.into("f_sources").columns("id", "uid")

    for source_uid in source_uids:
        query = query.insert(new_source_id(), source_uid)

    query = query.on_conflict("uid").do_nothing()

    await execute(str(query))

    result = await execute(
        "SELECT id, uid FROM f_sources WHERE uid = ANY(%(source_uids)s)", {"source_uids": list(source_uids)}
    )

    return {row["uid"]: row["id"] for row in result}


async def tech_remove_feed(feed_id: FeedId) -> None:
    sql = """
    DELETE FROM f_feeds
    WHERE id = %(feed_id)s
    """

    await execute(sql, {"feed_id": feed_id})


async def count_total_feeds() -> int:
    result = await execute("SELECT COUNT(*) FROM f_feeds")
    return result[0]["count"]  # type: ignore


async def count_total_feeds_per_state() -> dict[FeedState, int]:

    numbers: dict[FeedState, int] = {state: 0 for state in FeedState}

    result = await execute("SELECT state, COUNT(*) FROM f_feeds GROUP BY state")

    for row in result:
        numbers[FeedState(row["state"])] = row["count"]

    return numbers


async def count_total_feeds_per_last_error() -> dict[FeedError, int]:

    numbers: dict[FeedError, int] = {error: 0 for error in FeedError}

    result = await execute("SELECT last_error, COUNT(*) FROM f_feeds GROUP BY last_error")

    for row in result:
        if row["last_error"] is None:
            continue

        numbers[FeedError(row["last_error"])] = row["count"]

    return numbers



import pytest

from ffun.domain.datetime_intervals import month_interval_start
from ffun.domain.entities import UserId
from ffun.resources.domain import load_resource
from ffun.resources.operations import initialize_resource, try_to_reserve

_kind = 214
_another_kind = 215


class TestLoadResource:
    @pytest.mark.asyncio
    async def test_initialized(self, internal_user_id: UserId) -> None:
        interval_started_at = month_interval_start()

        await initialize_resource(user_id=internal_user_id, kind=_kind, interval_started_at=interval_started_at)

        await try_to_reserve(
            user_id=internal_user_id, kind=_kind, interval_started_at=interval_started_at, amount=13, limit=100
        )

        resource = await load_resource(user_id=internal_user_id, kind=_kind, interval_started_at=interval_started_at)

        assert resource.user_id == internal_user_id
        assert resource.kind == _kind
        assert resource.interval_started_at == interval_started_at
        assert resource.used == 0
        assert resource.reserved == 13

    @pytest.mark.asyncio
    async def test_not_initialized(self, internal_user_id: UserId) -> None:
        interval_started_at = month_interval_start()

        resource = await load_resource(user_id=internal_user_id, kind=_kind, interval_started_at=interval_started_at)

        assert resource.user_id == internal_user_id
        assert resource.kind == _kind
        assert resource.interval_started_at == interval_started_at
        assert resource.used == 0
        assert resource.reserved == 0



import uuid

import pytest
import pytest_asyncio

from ffun.domain.entities import TagId, TagUid
from ffun.ontology.domain import get_ids_by_uids
from ffun.ontology.entities import ProcessorTag


@pytest_asyncio.fixture
async def three_tags_by_uids() -> dict[TagUid, TagId]:
    return await get_ids_by_uids([uuid.uuid4().hex for _ in range(3)])  # type: ignore


@pytest.fixture
def three_tags_by_ids(three_tags_by_uids: dict[TagUid, TagId]) -> dict[TagId, TagUid]:
    return {v: k for k, v in three_tags_by_uids.items()}


@pytest.fixture
def three_tags_ids(three_tags_by_uids: dict[TagUid, TagId]) -> tuple[TagId, TagId, TagId]:
    return tuple(three_tags_by_uids.values())  # type: ignore


@pytest.fixture
def three_processor_tags(three_tags_by_ids: dict[TagId, TagUid]) -> tuple[ProcessorTag, ProcessorTag, ProcessorTag]:
    return tuple(ProcessorTag(raw_uid=tag) for tag in three_tags_by_ids.values())  # type: ignore


@pytest_asyncio.fixture
async def five_tags_by_uids() -> dict[TagUid, TagId]:
    return await get_ids_by_uids([uuid.uuid4().hex for _ in range(5)])  # type: ignore


@pytest.fixture
def five_tags_by_ids(five_tags_by_uids: dict[TagUid, TagId]) -> dict[TagId, TagUid]:
    return {v: k for k, v in five_tags_by_uids.items()}


@pytest.fixture
def five_tags_ids(five_tags_by_uids: dict[TagUid, TagId]) -> tuple[TagId, TagId, TagId, TagId, TagId]:
    return tuple(five_tags_by_uids.values())  # type: ignore


@pytest.fixture
def five_processor_tags(
    five_tags_by_ids: dict[TagId, TagUid],
) -> tuple[ProcessorTag, ProcessorTag, ProcessorTag, ProcessorTag, ProcessorTag]:
    return tuple(ProcessorTag(raw_uid=tag) for tag in five_tags_by_ids.values())  # type: ignore



import uuid

import pytest

from ffun.core.postgresql import execute
from ffun.core.tests.helpers import (
    Delta,
    TableSizeDelta,
    TableSizeNotChanged,
    assert_logs_has_business_event,
    assert_logs_has_no_business_event,
    capture_logs,
)
from ffun.core.utils import now
from ffun.domain.domain import new_user_id
from ffun.domain.entities import UserId
from ffun.users import errors
from ffun.users.entities import Service
from ffun.users.operations import (
    add_mapping,
    count_total_users,
    get_mapping,
    get_user_external_ids,
    store_user,
    unlink_user,
)


class TestAddMapping:

    @pytest.mark.asyncio
    async def test_new_user(self, external_user_id: str) -> None:
        with capture_logs() as logs:
            async with TableSizeDelta("u_users", delta=1):
                async with TableSizeDelta("u_mapping", delta=1):
                    internal_user_id = await add_mapping(Service.supertokens, external_user_id)

        assert await get_mapping(Service.supertokens, external_user_id) == internal_user_id

        assert_logs_has_business_event(logs, "user_created", user_id=internal_user_id)

    @pytest.mark.asyncio
    async def test_different_service(self, external_user_id: str) -> None:
        async with TableSizeDelta("u_users", delta=2):
            async with TableSizeDelta("u_mapping", delta=2):
                internal_user_id_1 = await add_mapping(Service.supertokens, external_user_id)
                internal_user_id_2 = await add_mapping(Service.single, external_user_id)

        assert await get_mapping(Service.supertokens, external_user_id) == internal_user_id_1
        assert await get_mapping(Service.single, external_user_id) == internal_user_id_2

    @pytest.mark.asyncio
    async def test_existing_user(self, external_user_id: str, internal_user_id: UserId) -> None:
        with capture_logs() as logs:
            async with TableSizeNotChanged("u_users"):
                async with TableSizeNotChanged("u_mapping"):
                    assert await add_mapping(Service.supertokens, external_user_id) == internal_user_id

        assert await get_mapping(Service.supertokens, external_user_id) == internal_user_id

        assert_logs_has_no_business_event(logs, "user_created")


# most of functionality are tested in other tests
class TestGetMapping:

    @pytest.mark.asyncio
    async def test_no_user_found(self, external_user_id: str) -> None:
        with pytest.raises(errors.NoUserMappingFound):
            await get_mapping(Service.supertokens, external_user_id)

    @pytest.mark.asyncio
    async def test_no_user_found_for_another_service(self, external_user_id: str) -> None:
        await add_mapping(Service.supertokens, external_user_id)

        with pytest.raises(errors.NoUserMappingFound):
            await get_mapping(Service.single, external_user_id)


class TestCountTotalUsers:

    @pytest.mark.asyncio
    async def test(self) -> None:
        async with Delta(count_total_users, delta=3):
            for _ in range(3):
                await add_mapping(Service.supertokens, uuid.uuid4().hex)


class TestStoreUser:

    @pytest.mark.asyncio
    async def test_new_user(self) -> None:
        user_id = new_user_id()

        async with TableSizeDelta("u_users", delta=1):
            await store_user(execute, user_id)

        result = await execute("SELECT * FROM u_users WHERE id = %(id)s", {"id": user_id})

        assert len(result) == 1
        assert result[0]["id"] == user_id

    @pytest.mark.asyncio
    async def test_existing_user(self) -> None:
        user_id = new_user_id()

        await store_user(execute, user_id)

        timestamp = now()

        async with TableSizeNotChanged("u_users"):
            await store_user(execute, user_id)

        result = await execute("SELECT * FROM u_users WHERE id = %(id)s", {"id": user_id})

        assert len(result) == 1
        assert result[0]["id"] == user_id
        assert result[0]["created_at"] < timestamp


# currently we do not support multiple mappings for the same user
# TODO: add tests for multiple mappings when we support it
class TestGetUserExternalIds:

    @pytest.mark.asyncio
    async def test_no_mappings(self) -> None:
        user_id = new_user_id()
        assert await get_user_external_ids(user_id) == {}

    @pytest.mark.asyncio
    async def test_single_mappings(self, internal_user_id: UserId, external_user_id: str) -> None:
        mapping = await get_user_external_ids(internal_user_id)

        assert mapping == {Service.supertokens: external_user_id}


# TODO: add tests for multiple mappings when we support it
class TestUnlinkUser:

    @pytest.mark.asyncio
    async def test_no_mappings(self) -> None:
        user_id = new_user_id()
        async with TableSizeNotChanged("u_mapping"):
            await unlink_user(Service.supertokens, user_id)

    @pytest.mark.asyncio
    async def test_single_mappings(self, internal_user_id: UserId) -> None:
        async with TableSizeNotChanged("u_users"):
            async with TableSizeDelta("u_mapping", delta=-1):
                await unlink_user(Service.supertokens, internal_user_id)

        mapping = await get_user_external_ids(internal_user_id)

        assert mapping == {}



import datetime
from typing import Any, Iterable

from ffun.core import logging
from ffun.core.postgresql import execute
from ffun.domain.entities import UserId
from ffun.resources import errors
from ffun.resources.entities import Resource

logger = logging.get_module_logger()


def row_to_entry(row: dict[str, Any]) -> Resource:
    return Resource(
        user_id=row["user_id"],
        kind=row["kind"],
        interval_started_at=row["interval_started_at"],
        used=row["used"],
        reserved=row["reserved"],
    )


async def initialize_resource(user_id: UserId, kind: int, interval_started_at: datetime.datetime) -> Resource:
    sql = """
        INSERT INTO r_resources (user_id, kind, interval_started_at)
        VALUES (%(user_id)s, %(kind)s, %(interval_started_at)s)
        ON CONFLICT (user_id, kind, interval_started_at) DO NOTHING
    """

    await execute(sql, {"user_id": user_id, "kind": kind, "interval_started_at": interval_started_at})

    sql = """
    SELECT * FROM r_resources
    WHERE user_id = %(user_id)s AND kind = %(kind)s AND interval_started_at = %(interval_started_at)s
    """

    results = await execute(sql, {"user_id": user_id, "kind": kind, "interval_started_at": interval_started_at})

    return row_to_entry(results[0])


async def load_resources(
    user_ids: Iterable[UserId], kind: int, interval_started_at: datetime.datetime
) -> dict[UserId, Resource]:
    sql = """
        SELECT * FROM r_resources
        WHERE user_id = ANY(%(user_ids)s) AND kind = %(kind)s AND interval_started_at = %(interval_started_at)s
    """

    results = await execute(
        sql, {"user_ids": list(user_ids), "kind": kind, "interval_started_at": interval_started_at}
    )

    resources = {}

    for row in results:
        resources[row["user_id"]] = row_to_entry(row)

    for user_id in user_ids:
        if user_id not in resources:
            resources[user_id] = await initialize_resource(user_id, kind, interval_started_at)

    return resources


async def try_to_reserve(
    user_id: UserId, kind: int, interval_started_at: datetime.datetime, amount: int, limit: int
) -> bool:
    await initialize_resource(user_id, kind, interval_started_at)

    sql = """
        UPDATE r_resources
        SET reserved = reserved + %(amount)s,
            updated_at = NOW()
        WHERE user_id = %(user_id)s AND
              kind = %(kind)s AND
              interval_started_at = %(interval_started_at)s AND
              used + reserved + %(amount)s <= %(limit)s
        RETURNING *
    """

    results = await execute(
        sql,
        {
            "user_id": user_id,
            "kind": kind,
            "interval_started_at": interval_started_at,
            "amount": amount,
            "limit": limit,
        },
    )

    return len(results) > 0


async def convert_reserved_to_used(
    user_id: UserId, kind: int, interval_started_at: datetime.datetime, used: int, reserved: int
) -> None:
    sql = """
        UPDATE r_resources
        SET used = used + %(used)s,
            reserved = reserved - %(reserved)s,
            updated_at = NOW()
        WHERE user_id = %(user_id)s AND
              kind = %(kind)s AND
              interval_started_at = %(interval_started_at)s AND
              reserved >= %(reserved)s
        RETURNING *
    """

    result = await execute(
        sql,
        {
            "user_id": user_id,
            "kind": kind,
            "interval_started_at": interval_started_at,
            "used": used,
            "reserved": reserved,
        },
    )

    if not result:
        raise errors.CanNotConvertReservedToUsed()


async def load_resource_history(user_id: UserId, kind: int) -> list[Resource]:
    sql = """
        SELECT * FROM r_resources
        WHERE user_id = %(user_id)s AND kind = %(kind)s
        ORDER BY interval_started_at DESC
    """

    results = await execute(sql, {"user_id": user_id, "kind": kind})

    return [row_to_entry(row) for row in results]


async def count_total_resources_per_user(kind: int) -> dict[UserId, int]:
    sql = """
        SELECT user_id, SUM(used) AS count
        FROM r_resources
        WHERE kind = %(kind)s
        GROUP BY user_id
    """

    results = await execute(sql, {"kind": kind})

    return {row["user_id"]: row["count"] for row in results}



from typing import Iterable

from ffun.domain.entities import UserId
from ffun.user_settings import operations


async def has_settings(user_id: UserId, values: dict[int, str]) -> None:
    settings = await operations.load_settings_for_users([user_id], list(values.keys()))
    assert settings[user_id] == values


async def has_no_settings(user_id: UserId, kinds: Iterable[int]) -> None:
    settings = await operations.load_settings_for_users([user_id], list(kinds))
    assert set(kinds) - set(settings[user_id]) == set(kinds)



"""
fill-uids-for-feeds
"""

from typing import Any

from psycopg import Connection
from yoyo import step

__depends__ = {"20240504_01_vBDVJ-column-for-feed-unique-id"}


def get_feeds(cursor: Any) -> list[dict[str, Any]]:
    cursor.execute("SELECT id, url FROM f_feeds")

    return [{"id": row[0], "url": row[1]} for row in cursor.fetchall()]


def apply_step(conn: Connection[dict[str, Any]]) -> None:
    pass


def rollback_step(conn: Connection[dict[str, Any]]) -> None:
    pass


steps = [step(apply_step, rollback_step)]



from typing import Any

from psycopg import Connection
from yoyo import step

__depends__: set[str] = set()

sql_create_tags_table = """
CREATE TABLE o_tags (
    id BIGSERIAL PRIMARY KEY,
    name TEXT NOT NULL UNIQUE,
    created_at TIMESTAMP WITH TIME ZONE NOT NULL DEFAULT NOW())
"""

sql_create_relations_table = """
CREATE TABLE o_relations (
    id BIGSERIAL PRIMARY KEY,
    entry_id UUID NOT NULL,
    tag_id BIGINT NOT NULL REFERENCES o_tags(id),
    created_at TIMESTAMP WITH TIME ZONE NOT NULL DEFAULT NOW())
"""

sql_create_relations_entry_id_tag_id_index = """
CREATE UNIQUE INDEX idx_o_relations_entry_id_tag_id ON o_relations (entry_id, tag_id);
"""


def apply_step(conn: Connection[dict[str, Any]]) -> None:
    cursor = conn.cursor()
    cursor.execute(sql_create_tags_table)
    cursor.execute(sql_create_relations_table)
    cursor.execute(sql_create_relations_entry_id_tag_id_index)


def rollback_step(conn: Connection[dict[str, Any]]) -> None:
    cursor = conn.cursor()
    cursor.execute("DROP TABLE o_tags, o_relations")


steps = [step(apply_step, rollback_step)]



import datetime
from typing import Iterable

from ffun.core.postgresql import execute
from ffun.domain.entities import EntryId


async def update_cataloged_time(entries_ids: Iterable[EntryId], new_time: datetime.datetime) -> None:
    await execute(
        "UPDATE l_feeds_to_entries SET created_at = %(time_border)s WHERE entry_id = ANY(%(ids)s)",
        {
            "time_border": new_time,
            "ids": list(entries_ids),
        },
    )

    await execute(
        "UPDATE l_entries SET created_at = %(time_border)s WHERE id = ANY(%(ids)s)",
        {
            "time_border": new_time,
            "ids": list(entries_ids),
        },
    )



import datetime

import pytest

from ffun.core.tests.helpers import TableSizeDelta, TableSizeNotChanged
from ffun.domain.datetime_intervals import month_interval_start
from ffun.domain.entities import UserId
from ffun.resources import errors
from ffun.resources.domain import load_resource
from ffun.resources.operations import (
    convert_reserved_to_used,
    count_total_resources_per_user,
    initialize_resource,
    load_resource_history,
    load_resources,
    try_to_reserve,
)


@pytest.fixture
def interval_started_at() -> datetime.datetime:
    return month_interval_start()


_kind = 214
_another_kind = 215


class TestInitializeResource:
    @pytest.mark.asyncio
    async def test_new_record(self, internal_user_id: UserId, interval_started_at: datetime.datetime) -> None:
        async with TableSizeDelta("r_resources", delta=1):
            resource = await initialize_resource(
                user_id=internal_user_id, kind=_kind, interval_started_at=interval_started_at
            )

        assert resource.user_id == internal_user_id
        assert resource.kind == _kind
        assert resource.interval_started_at == interval_started_at
        assert resource.used == 0
        assert resource.reserved == 0

        async with TableSizeNotChanged("r_resources"):
            loaded_resource = await load_resource(
                user_id=internal_user_id, kind=_kind, interval_started_at=interval_started_at
            )

        assert loaded_resource == resource

    @pytest.mark.asyncio
    async def test_do_not_reinitialized_if_exists(
        self, internal_user_id: UserId, interval_started_at: datetime.datetime
    ) -> None:
        await initialize_resource(user_id=internal_user_id, kind=_kind, interval_started_at=interval_started_at)

        await try_to_reserve(
            user_id=internal_user_id, kind=_kind, interval_started_at=interval_started_at, amount=1, limit=100
        )

        async with TableSizeNotChanged("r_resources"):
            resource = await initialize_resource(
                user_id=internal_user_id, kind=_kind, interval_started_at=interval_started_at
            )

        assert resource.reserved == 1

        loaded_resource = await load_resource(
            user_id=internal_user_id, kind=_kind, interval_started_at=interval_started_at
        )

        assert loaded_resource == resource


class TestLoadResources:
    """Most functionality are tested in other classes."""

    @pytest.mark.asyncio
    async def test_initialize_if_not_found(
        self, internal_user_id: UserId, another_internal_user_id: UserId, interval_started_at: datetime.datetime
    ) -> None:
        await initialize_resource(user_id=internal_user_id, kind=_kind, interval_started_at=interval_started_at)

        await try_to_reserve(
            user_id=internal_user_id, kind=_kind, interval_started_at=interval_started_at, amount=13, limit=100
        )

        async with TableSizeDelta("r_resources", delta=1):
            resources = await load_resources(
                user_ids=[internal_user_id, another_internal_user_id],
                kind=_kind,
                interval_started_at=interval_started_at,
            )

        assert len(resources) == 2

        resource_1 = resources[internal_user_id]

        assert resource_1.user_id == internal_user_id
        assert resource_1.kind == _kind
        assert resource_1.interval_started_at == interval_started_at
        assert resource_1.used == 0
        assert resource_1.reserved == 13

        resource_2 = resources[another_internal_user_id]

        assert resource_2.user_id == another_internal_user_id
        assert resource_2.kind == _kind
        assert resource_2.interval_started_at == interval_started_at
        assert resource_2.used == 0
        assert resource_2.reserved == 0


class TestTryToReserve:
    @pytest.mark.parametrize("amount", [0, 1, 100])
    @pytest.mark.asyncio
    async def test_for_not_existed_resource(
        self, amount: int, internal_user_id: UserId, interval_started_at: datetime.datetime
    ) -> None:
        result = await try_to_reserve(
            user_id=internal_user_id, kind=_kind, interval_started_at=interval_started_at, amount=amount, limit=100
        )

        assert result

        resource = await load_resource(user_id=internal_user_id, kind=_kind, interval_started_at=interval_started_at)

        assert resource.used == 0
        assert resource.reserved == amount

    @pytest.mark.asyncio
    async def test_for_existed_resource(
        self, internal_user_id: UserId, interval_started_at: datetime.datetime
    ) -> None:
        result = await try_to_reserve(
            user_id=internal_user_id, kind=_kind, interval_started_at=interval_started_at, amount=1, limit=100
        )

        result = await try_to_reserve(
            user_id=internal_user_id, kind=_kind, interval_started_at=interval_started_at, amount=13, limit=100
        )

        assert result

        resource = await load_resource(user_id=internal_user_id, kind=_kind, interval_started_at=interval_started_at)

        assert resource.used == 0
        assert resource.reserved == 14

    @pytest.mark.asyncio
    async def test_not_enough(self, internal_user_id: UserId, interval_started_at: datetime.datetime) -> None:
        result = await try_to_reserve(
            user_id=internal_user_id, kind=_kind, interval_started_at=interval_started_at, amount=101, limit=100
        )

        assert not result


class TestConvertReservedToUsed:
    @pytest.mark.parametrize(
        "reserved, converted_reserved, converted_used, expected_reserved, expected_used",
        [(13, 13, 9, 0, 9), (13, 13, 13, 0, 13), (77, 14, 13, 63, 13)],
    )
    @pytest.mark.asyncio
    async def test_converted(  # noqa: CFQ002
        self,
        internal_user_id: UserId,
        interval_started_at: datetime.datetime,
        reserved: int,
        converted_reserved: int,
        converted_used: int,
        expected_reserved: int,
        expected_used: int,
    ) -> None:
        await try_to_reserve(
            user_id=internal_user_id, kind=_kind, interval_started_at=interval_started_at, amount=reserved, limit=100
        )

        await convert_reserved_to_used(
            user_id=internal_user_id,
            kind=_kind,
            interval_started_at=interval_started_at,
            reserved=converted_reserved,
            used=converted_used,
        )

        resource = await load_resource(user_id=internal_user_id, kind=_kind, interval_started_at=interval_started_at)

        assert resource.used == expected_used
        assert resource.reserved == expected_reserved

    @pytest.mark.asyncio
    async def test_not_enough(self, internal_user_id: UserId, interval_started_at: datetime.datetime) -> None:
        await try_to_reserve(
            user_id=internal_user_id, kind=_kind, interval_started_at=interval_started_at, amount=13, limit=100
        )

        with pytest.raises(errors.CanNotConvertReservedToUsed):
            await convert_reserved_to_used(
                user_id=internal_user_id, kind=_kind, interval_started_at=interval_started_at, reserved=14, used=13
            )

    @pytest.mark.asyncio
    async def test_no_resource(self, internal_user_id: UserId, interval_started_at: datetime.datetime) -> None:
        with pytest.raises(errors.CanNotConvertReservedToUsed):
            await convert_reserved_to_used(
                user_id=internal_user_id, kind=_kind, interval_started_at=interval_started_at, reserved=0, used=13
            )


class TestLoadResourceHistory:
    @pytest.mark.asyncio
    async def test_no_history(self, internal_user_id: UserId) -> None:
        history = await load_resource_history(user_id=internal_user_id, kind=_kind)

        assert len(history) == 0

    @pytest.mark.asyncio
    async def test_with_history(self, internal_user_id: UserId, another_internal_user_id: UserId) -> None:
        internal_1 = datetime.datetime(2020, 1, 1, 0, 0, 0, tzinfo=datetime.timezone.utc)
        internal_2 = datetime.datetime(2020, 2, 1, 0, 0, 0, tzinfo=datetime.timezone.utc)
        internal_3 = datetime.datetime(2020, 3, 1, 0, 0, 0, tzinfo=datetime.timezone.utc)

        await try_to_reserve(
            user_id=internal_user_id, kind=_kind, interval_started_at=internal_1, amount=13, limit=100
        )

        await try_to_reserve(
            user_id=internal_user_id, kind=_kind, interval_started_at=internal_3, amount=14, limit=100
        )

        await try_to_reserve(
            user_id=another_internal_user_id, kind=_kind, interval_started_at=internal_2, amount=15, limit=100
        )

        await try_to_reserve(
            user_id=internal_user_id, kind=_another_kind, interval_started_at=internal_3, amount=16, limit=100
        )

        history = await load_resource_history(user_id=internal_user_id, kind=_kind)

        assert len(history) == 2

        assert history[0].user_id == internal_user_id
        assert history[0].interval_started_at == internal_3
        assert history[0].reserved == 14

        assert history[1].user_id == internal_user_id
        assert history[1].interval_started_at == internal_1
        assert history[1].reserved == 13

        history = await load_resource_history(user_id=another_internal_user_id, kind=_kind)

        assert len(history) == 1

        assert history[0].user_id == another_internal_user_id
        assert history[0].interval_started_at == internal_2
        assert history[0].reserved == 15


class TestCountTotalResourcesPerUser:

    @pytest.mark.asyncio
    async def test(self, internal_user_id: UserId, another_internal_user_id: UserId) -> None:
        await try_to_reserve(
            user_id=internal_user_id,
            kind=_kind,
            interval_started_at=datetime.datetime(2020, 1, 1, 0, 0, 0, tzinfo=datetime.timezone.utc),
            amount=13,
            limit=100,
        )
        await convert_reserved_to_used(
            user_id=internal_user_id,
            kind=_kind,
            interval_started_at=datetime.datetime(2020, 1, 1, 0, 0, 0, tzinfo=datetime.timezone.utc),
            reserved=10,
            used=10,
        )

        await try_to_reserve(
            user_id=internal_user_id,
            kind=_kind,
            interval_started_at=datetime.datetime(2020, 3, 1, 0, 0, 0, tzinfo=datetime.timezone.utc),
            amount=14,
            limit=100,
        )
        await convert_reserved_to_used(
            user_id=internal_user_id,
            kind=_kind,
            interval_started_at=datetime.datetime(2020, 3, 1, 0, 0, 0, tzinfo=datetime.timezone.utc),
            reserved=14,
            used=14,
        )

        await try_to_reserve(
            user_id=internal_user_id,
            kind=_another_kind,
            interval_started_at=datetime.datetime(2020, 3, 1, 0, 0, 0, tzinfo=datetime.timezone.utc),
            amount=6,
            limit=100,
        )
        await convert_reserved_to_used(
            user_id=internal_user_id,
            kind=_another_kind,
            interval_started_at=datetime.datetime(2020, 3, 1, 0, 0, 0, tzinfo=datetime.timezone.utc),
            reserved=6,
            used=6,
        )

        await try_to_reserve(
            user_id=another_internal_user_id,
            kind=_kind,
            interval_started_at=datetime.datetime(2020, 2, 1, 0, 0, 0, tzinfo=datetime.timezone.utc),
            amount=15,
            limit=100,
        )
        await convert_reserved_to_used(
            user_id=another_internal_user_id,
            kind=_kind,
            interval_started_at=datetime.datetime(2020, 2, 1, 0, 0, 0, tzinfo=datetime.timezone.utc),
            reserved=14,
            used=14,
        )

        await try_to_reserve(
            user_id=another_internal_user_id,
            kind=_another_kind,
            interval_started_at=datetime.datetime(2020, 2, 1, 0, 0, 0, tzinfo=datetime.timezone.utc),
            amount=6,
            limit=100,
        )
        await convert_reserved_to_used(
            user_id=another_internal_user_id,
            kind=_another_kind,
            interval_started_at=datetime.datetime(2020, 2, 1, 0, 0, 0, tzinfo=datetime.timezone.utc),
            reserved=6,
            used=6,
        )

        numbers = await count_total_resources_per_user(kind=_kind)

        assert numbers[internal_user_id] == 24
        assert numbers[another_internal_user_id] == 14



"""
Development utilities for feeds.

This code is intended to be used by developers to simplify our work, therefore:

- It MUST NOT be used in production code.
- It MUST NOT be used in tests.
- no tests are required for this code.
"""

import asyncio
import uuid

import typer

from ffun.application.application import with_app
from ffun.auth.settings import settings as a_settings
from ffun.core import logging, utils
from ffun.domain.domain import new_entry_id, new_feed_id
from ffun.domain.entities import UnknownUrl, UserId
from ffun.domain.urls import adjust_classic_relative_url, str_to_feed_url, url_to_source_uid
from ffun.feeds.domain import get_feeds, get_source_ids, save_feed
from ffun.feeds.entities import Feed, FeedState
from ffun.feeds_links.domain import add_link
from ffun.library.domain import catalog_entries, get_entries_by_ids
from ffun.library.entities import Entry
from ffun.ontology.domain import apply_tags_to_entry
from ffun.ontology.entities import ProcessorTag
from ffun.users import domain as u_domain
from ffun.users import entities as u_entities

logger = logging.get_module_logger()

cli_app = typer.Typer()


async def fake_feed() -> Feed:

    _id = uuid.uuid4().hex

    url = str_to_feed_url(f"https://{_id}.com")

    source_uid = url_to_source_uid(url)

    source_ids = await get_source_ids([source_uid])

    timestamp = utils.now()

    feed = Feed(
        id=new_feed_id(),
        source_id=source_ids[source_uid],
        url=url,
        state=FeedState.loaded,
        last_error=None,
        load_attempted_at=timestamp,
        loaded_at=timestamp,
        title=f"Title {_id}",
        description=f"Description {_id}",
    )

    feed_id = await save_feed(feed)

    feeds = await get_feeds([feed_id])

    return feeds[0]


async def fake_entry(feed: Feed) -> Entry:
    _id = uuid.uuid4().hex

    timestamp = utils.now()

    url = adjust_classic_relative_url(UnknownUrl(f"enrty-{_id}"), feed.url)

    assert url is not None

    entry = Entry(
        id=new_entry_id(),
        source_id=feed.source_id,
        title=f"Title {_id}",
        body=f"Body {_id}",
        external_id=uuid.uuid4().hex,
        external_url=url,
        external_tags=set(),
        published_at=timestamp,
        cataloged_at=timestamp,
    )

    await catalog_entries(feed.id, [entry])

    entries = await get_entries_by_ids([entry.id])

    assert entry is not None

    returned_entry = entries[entry.id]

    assert returned_entry is not None

    return returned_entry


async def run_fill_db(feeds_number: int, entries_per_feed: int, tags_per_entry: int) -> None:
    async with with_app():
        external_user_id = a_settings.single_user.external_id

        internal_user_id = await u_domain.get_or_create_user_id(u_entities.Service.single, external_user_id)

        for _ in range(feeds_number):
            feed = await fake_feed()

            await add_link(internal_user_id, feed.id)

            entries = [await fake_entry(feed) for _ in range(entries_per_feed)]

            for i, entry in enumerate(entries, start=1):

                tags = []

                for j, _ in enumerate(range(tags_per_entry), start=1):
                    raw_uid = f"some-long-tag-name-{j}-{i % j}"
                    tags.append(ProcessorTag(raw_uid=raw_uid))

                await apply_tags_to_entry(entry.id, 100500, tags)


@cli_app.command()
def fill_db(feeds_number: int = 10, entries_per_feed: int = 100, tags_per_entry: int = 25) -> None:
    asyncio.run(run_fill_db(feeds_number, entries_per_feed, tags_per_entry))


async def run_supertokens_user_to_dev(intenal_user_id: UserId) -> None:
    async with with_app():
        external_user_id = a_settings.single_user.external_id

        to_user_id = await u_domain.get_or_create_user_id(u_entities.Service.single, external_user_id)

        await u_domain.tech_move_user(from_user_id=intenal_user_id, to_user_id=to_user_id)


@cli_app.command()
def supertokens_user_to_dev(intenal_user_id: str) -> None:
    asyncio.run(run_supertokens_user_to_dev(UserId(uuid.UUID(intenal_user_id))))



import decimal

from ffun.core.register import Entity, Register
from ffun.user_settings.types import Type


class Value(Entity):
    def __init__(
        self,
        key: int,
        name: str,
        type: Type,
        default: str | int | bool | decimal.Decimal | None,
    ) -> None:
        super().__init__(key=key)
        self.name = name
        self.type = type
        self.default = default


SettingsRegister = Register[Value]


user_settings: SettingsRegister = Register()



import datetime

from ffun.domain.entities import UserId
from ffun.resources import operations
from ffun.resources.entities import Resource

load_resources = operations.load_resources
try_to_reserve = operations.try_to_reserve
convert_reserved_to_used = operations.convert_reserved_to_used
load_resource_history = operations.load_resource_history
count_total_resources_per_user = operations.count_total_resources_per_user


async def load_resource(user_id: UserId, kind: int, interval_started_at: datetime.datetime) -> Resource:
    resources = await load_resources([user_id], kind, interval_started_at)
    return resources[user_id]



from importlib.metadata import version

from ffun.application.settings import settings


def user_agent() -> str:
    name = settings.app_name.replace(" ", "")
    package_version = version("ffun")
    return f"{name}/{package_version} ({settings.environment} {settings.app_domain})"



"""
fix-unique-index-on-l-entries
"""

from typing import Any

from psycopg import Connection
from psycopg.rows import dict_row
from yoyo import step

__depends__ = {"20240506_01_My4vi-remove-duplicated-entries-from-feeds"}


index = "l_entries_feed_id_external_id_key"


def apply_step(conn: Connection[dict[str, Any]]) -> None:
    cursor = conn.cursor(row_factory=dict_row)

    cursor.execute(f"DROP INDEX {index}")
    cursor.execute(f"CREATE UNIQUE INDEX {index} ON l_entries (feed_id, external_id)")


def rollback_step(conn: Connection[dict[str, Any]]) -> None:
    cursor = conn.cursor(row_factory=dict_row)

    cursor.execute(f"DROP INDEX {index}")
    cursor.execute(f"CREATE INDEX {index} ON l_entries (feed_id, external_id)")


steps = [step(apply_step, rollback_step)]



from ffun.auth import supertokens
from ffun.users import domain as u_domain
from ffun.users import entities as u_entities


async def remove_user_from_external_service(service: u_entities.Service, external_user_id: str) -> None:
    match service:
        case u_entities.Service.supertokens:
            await supertokens.remove_user(external_user_id)
        case u_entities.Service.single:
            # do nothing because we have no external service in this case
            pass


async def logout_user_from_all_sessions_in_service(service: u_entities.Service, external_user_id: str) -> None:
    match service:
        case u_entities.Service.supertokens:
            await supertokens.logout_user_from_all_sessions(external_user_id)
        case u_entities.Service.single:
            # do nothing because user is always logged in in this case
            pass


async def logout_user_from_all_sessions(user_id: u_entities.UserId) -> None:
    external_ids = await u_domain.get_user_external_ids(user_id)

    for service, external_id in external_ids.items():
        await logout_user_from_all_sessions_in_service(service, external_id)



from ffun.core import logging
from ffun.librarian.processors import base
from ffun.library.entities import Entry
from ffun.ontology.entities import ProcessorTag

logger = logging.get_module_logger()


class Processor(base.Processor):
    __slots__ = ()

    async def process(self, entry: Entry) -> list[ProcessorTag]:
        if entry.title.isupper():
            return [ProcessorTag(raw_uid="upper-case-title")]

        return []



import copy
import enum
import uuid
from typing import Any

import pytest

from ffun.core.register import Register
from ffun.core.tests.helpers import TableSizeDecreased, TableSizeNotChanged
from ffun.domain.entities import UserId
from ffun.user_settings import domain, errors, operations, types
from ffun.user_settings.tests import asserts
from ffun.user_settings.values import SettingsRegister, Value

_kind_1 = 1234
_kind_2 = 5678


class Setting(enum.IntEnum):
    kind_integer = 1
    kind_string = 2
    kind_boolean = 3
    kind_secret = 4


_integer_default = 123
_string_default = "abc"
_boolean_default = True
_secret_default = "secret"  # noqa: S105


register: SettingsRegister = Register()


register.add(Value(key=Setting.kind_integer, name="integer", type=types.Integer(), default=_integer_default))

register.add(Value(key=Setting.kind_string, name="string", type=types.String(), default=_string_default))

register.add(Value(key=Setting.kind_boolean, name="boolean", type=types.Boolean(), default=_boolean_default))

register.add(Value(key=Setting.kind_secret, name="secret", type=types.Secret(), default=_secret_default))


class TestSave:
    @pytest.mark.asyncio
    async def test_save_with_conversion(self, internal_user_id: UserId) -> None:
        await domain.save_setting(internal_user_id, Setting.kind_integer, 124, register=register)
        await asserts.has_settings(internal_user_id, {Setting.kind_integer: "124"})

    @pytest.mark.asyncio
    async def test_do_not_save_after_error(self, internal_user_id: UserId) -> None:
        with pytest.raises(errors.WrongValueType):
            await domain.save_setting(
                internal_user_id, Setting.kind_integer, "string instead of int", register=register
            )

        await asserts.has_no_settings(internal_user_id, {Setting.kind_integer})


class TestFullSettings:
    def test(self) -> None:
        values: dict[int, Any] = {Setting.kind_integer: "666", Setting.kind_boolean: "false"}

        settings = domain._full_settings(values, kinds=list(Setting), register=register)

        assert settings == {
            Setting.kind_integer: 666,
            Setting.kind_boolean: False,
            Setting.kind_string: _string_default,
            Setting.kind_secret: _secret_default,
        }

    def test_skip_unknown(self) -> None:
        values: dict[int, Any] = {Setting.kind_integer: "666", _kind_1: "123"}

        settings = domain._full_settings(values, kinds=list(Setting) + [_kind_1], register=register)

        assert settings == {
            Setting.kind_integer: 666,
            Setting.kind_string: _string_default,
            Setting.kind_boolean: _boolean_default,
            Setting.kind_secret: _secret_default,
        }


class TestLoadSettings:
    @pytest.mark.asyncio
    async def test_no_settings(self, internal_user_id: UserId) -> None:
        settings = await domain.load_settings(internal_user_id, kinds=list(Setting), register=register)

        assert settings == {
            Setting.kind_integer: _integer_default,
            Setting.kind_string: _string_default,
            Setting.kind_boolean: _boolean_default,
            Setting.kind_secret: _secret_default,
        }

    @pytest.mark.asyncio
    async def test_has_settings(self, internal_user_id: UserId, another_internal_user_id: UserId) -> None:
        await domain.save_setting(internal_user_id, Setting.kind_integer, 124, register=register)
        await domain.save_setting(internal_user_id, Setting.kind_string, "xxxyyy", register=register)
        await domain.save_setting(another_internal_user_id, Setting.kind_integer, 421, register=register)

        settings = await domain.load_settings(internal_user_id, kinds=list(Setting), register=register)
        assert settings == {
            Setting.kind_integer: 124,
            Setting.kind_string: "xxxyyy",
            Setting.kind_boolean: _boolean_default,
            Setting.kind_secret: _secret_default,
        }

        settings = await domain.load_settings(another_internal_user_id, kinds=list(Setting), register=register)
        assert settings == {
            Setting.kind_integer: 421,
            Setting.kind_string: _string_default,
            Setting.kind_boolean: _boolean_default,
            Setting.kind_secret: _secret_default,
        }

    @pytest.mark.asyncio
    async def test_filter_by_kinds(self, internal_user_id: UserId, another_internal_user_id: UserId) -> None:
        await domain.save_setting(internal_user_id, Setting.kind_integer, 124, register=register)
        await domain.save_setting(internal_user_id, Setting.kind_string, "xxxyyy", register=register)

        settings = await domain.load_settings(
            internal_user_id, kinds=[Setting.kind_integer, Setting.kind_secret], register=register
        )
        assert settings == {Setting.kind_integer: 124, Setting.kind_secret: _secret_default}


class TestLoadSettingsForUsers:
    @pytest.mark.asyncio
    async def test(self, internal_user_id: UserId, another_internal_user_id: UserId) -> None:
        await domain.save_setting(internal_user_id, Setting.kind_integer, 124, register=register)
        await domain.save_setting(internal_user_id, Setting.kind_string, "xxxyyy", register=register)
        await domain.save_setting(another_internal_user_id, Setting.kind_integer, 421, register=register)
        await domain.save_setting(another_internal_user_id, Setting.kind_secret, "my-secret", register=register)

        kinds = [Setting.kind_integer, Setting.kind_secret, Setting.kind_string]

        settings = await domain.load_settings(internal_user_id, kinds=kinds, register=register)
        assert settings == {
            Setting.kind_integer: 124,
            Setting.kind_string: "xxxyyy",
            Setting.kind_secret: _secret_default,
        }

        settings = await domain.load_settings(another_internal_user_id, kinds=kinds, register=register)
        assert settings == {
            Setting.kind_integer: 421,
            Setting.kind_string: _string_default,
            Setting.kind_secret: "my-secret",
        }


class TestGetUsersWithSetting:
    @pytest.mark.asyncio
    async def test_wrong_value(self) -> None:
        with pytest.raises(errors.WrongValueType):
            await domain.get_users_with_setting(Setting.kind_integer, "sdassd", register=register)

    @pytest.mark.asyncio
    async def test_no_users(self) -> None:
        value = uuid.uuid4().int
        user_ids = await domain.get_users_with_setting(Setting.kind_integer, value, register=register)
        assert user_ids == set()

    @pytest.mark.asyncio
    async def test_found_users(self, internal_user_id: UserId, another_internal_user_id: UserId) -> None:
        value = uuid.uuid4().hex

        await domain.save_setting(internal_user_id, Setting.kind_string, value, register=register)
        await domain.save_setting(another_internal_user_id, Setting.kind_string, uuid.uuid4().hex, register=register)

        user_ids = await domain.get_users_with_setting(Setting.kind_string, value, register=register)

        assert user_ids == {internal_user_id}

        await domain.save_setting(another_internal_user_id, Setting.kind_string, value, register=register)

        user_ids = await domain.get_users_with_setting(Setting.kind_string, value, register=register)

        assert user_ids == {internal_user_id, another_internal_user_id}


class TestRemoveDeprecatedSettings:

    @pytest.mark.asyncio
    async def test_nothing_to_remove(self, internal_user_id: UserId, another_internal_user_id: UserId) -> None:
        # cleanup table for this test
        await domain.remove_deprecated_settings(register=register)

        str_values = [uuid.uuid4().hex, uuid.uuid4().hex]
        int_values = [uuid.uuid4().int, uuid.uuid4().int]

        await domain.save_setting(internal_user_id, Setting.kind_string, str_values[0], register=register)
        await domain.save_setting(another_internal_user_id, Setting.kind_string, str_values[1], register=register)

        await domain.save_setting(internal_user_id, Setting.kind_integer, int_values[0], register=register)
        await domain.save_setting(another_internal_user_id, Setting.kind_integer, int_values[1], register=register)

        await operations.find_all_kinds()

        async with TableSizeNotChanged("us_settings"):
            await domain.remove_deprecated_settings(register=register)

        user_ids = await domain.load_settings_for_users(
            [internal_user_id, another_internal_user_id],
            {Setting.kind_string, Setting.kind_integer},
            register=register,
        )

        assert user_ids == {
            internal_user_id: {Setting.kind_string: str_values[0], Setting.kind_integer: int_values[0]},
            another_internal_user_id: {Setting.kind_string: str_values[1], Setting.kind_integer: int_values[1]},
        }

    @pytest.mark.asyncio
    async def test_remove(self, internal_user_id: UserId, another_internal_user_id: UserId) -> None:
        str_values = [uuid.uuid4().hex, uuid.uuid4().hex]
        int_values = [uuid.uuid4().int, uuid.uuid4().int]

        await domain.save_setting(internal_user_id, Setting.kind_string, str_values[0], register=register)
        await domain.save_setting(another_internal_user_id, Setting.kind_string, str_values[1], register=register)

        await domain.save_setting(internal_user_id, Setting.kind_integer, int_values[0], register=register)
        await domain.save_setting(another_internal_user_id, Setting.kind_integer, int_values[1], register=register)

        await operations.find_all_kinds()

        updated_register = copy.deepcopy(register)
        updated_register.remove(Setting.kind_string)

        async with TableSizeDecreased("us_settings"):
            await domain.remove_deprecated_settings(register=updated_register)

        user_ids = await domain.load_settings_for_users(
            [internal_user_id, another_internal_user_id],
            {Setting.kind_string, Setting.kind_integer},
            register=updated_register,
        )

        assert user_ids == {
            internal_user_id: {Setting.kind_integer: int_values[0]},
            another_internal_user_id: {Setting.kind_integer: int_values[1]},
        }



import datetime

from ffun.core import utils


def month_interval_start(now: datetime.datetime | None = None) -> datetime.datetime:
    if now is None:
        now = utils.now()

    return now.replace(day=1, hour=0, minute=0, second=0, microsecond=0)



import datetime
import enum
from decimal import Decimal
from typing import Any, Iterable

import pydantic

from ffun.api import front_events
from ffun.core import api
from ffun.core.entities import BaseEntity
from ffun.domain.entities import (
    AbsoluteUrl,
    CollectionId,
    CollectionSlug,
    EntryId,
    FeedId,
    FeedUrl,
    RuleId,
    TagId,
    TagUid,
    UnknownUrl,
    UserId,
)
from ffun.feeds import entities as f_entities
from ffun.feeds_collections import entities as fc_entities
from ffun.feeds_links import entities as fl_entities
from ffun.library import entities as l_entities

# TODO: rename to public name
from ffun.llms_framework.keys_rotator import USDCost, _cost_points
from ffun.markers import entities as m_entities
from ffun.ontology import entities as o_entities
from ffun.parsers import entities as p_entities
from ffun.resources import entities as r_entities
from ffun.scores import entities as s_entities
from ffun.user_settings import types as us_types
from ffun.user_settings.values import user_settings


class Marker(enum.StrEnum):
    read = "read"

    @classmethod
    def from_internal(cls, marker: m_entities.Marker) -> "Marker":
        return cls(marker.name)

    def to_internal(self) -> m_entities.Marker:
        return m_entities.Marker[self.value]


class Feed(BaseEntity):
    id: FeedId
    title: str | None
    description: str | None
    url: FeedUrl
    state: str
    lastError: str | None = None
    loadedAt: datetime.datetime | None
    linkedAt: datetime.datetime | None
    collectionIds: list[CollectionId]

    @classmethod
    def from_internal(
        cls, feed: f_entities.Feed, link: fl_entities.FeedLink, collection_ids: list[CollectionId]
    ) -> "Feed":
        return cls(
            id=feed.id,
            title=feed.title,
            description=feed.description,
            url=feed.url,
            state=feed.state.name,
            lastError=feed.last_error.name if feed.last_error else None,
            loadedAt=feed.loaded_at,
            linkedAt=link.created_at,
            collectionIds=collection_ids,
        )


class Entry(BaseEntity):
    id: EntryId
    title: str
    url: AbsoluteUrl
    tags: list[TagId]
    markers: list[Marker] = []
    score: int
    scoreContributions: dict[TagId, int]
    publishedAt: datetime.datetime
    catalogedAt: datetime.datetime
    body: str | None = None

    @classmethod
    def from_internal(  # noqa: CFQ002
        cls,
        entry: l_entities.Entry,
        tags: Iterable[TagId],
        markers: Iterable[Marker],
        score: int,
        score_contributions: dict[TagId, int],
        with_body: bool = False,
    ) -> "Entry":
        return cls(
            id=entry.id,
            title=entry.title,
            url=entry.external_url,
            tags=list(tags),
            markers=list(markers),
            score=score,
            scoreContributions=score_contributions,
            publishedAt=entry.published_at,
            catalogedAt=entry.cataloged_at,
            body=entry.body if with_body else None,
        )


class Rule(BaseEntity):
    id: RuleId
    requiredTags: list[TagUid]
    excludedTags: list[TagUid]
    score: int
    createdAt: datetime.datetime
    updatedAt: datetime.datetime

    @classmethod
    def from_internal(cls, rule: s_entities.Rule, tags_mapping: dict[TagId, TagUid]) -> "Rule":
        return cls(
            id=rule.id,
            requiredTags=[tags_mapping[tag_id] for tag_id in rule.required_tags],
            excludedTags=[tags_mapping[tag_id] for tag_id in rule.excluded_tags],
            score=rule.score,
            createdAt=rule.created_at,
            updatedAt=rule.updated_at,
        )


class EntryInfo(BaseEntity):
    title: str
    body: str
    url: AbsoluteUrl | None
    published_at: datetime.datetime

    @classmethod
    def from_internal(cls, entry: p_entities.EntryInfo) -> "EntryInfo":
        return cls(title=entry.title, body=entry.body, url=entry.external_url, published_at=entry.published_at)


class FeedInfo(BaseEntity):
    url: FeedUrl
    title: str
    description: str
    isLinked: bool

    entries: list[EntryInfo]

    @classmethod
    def from_internal(cls, feed: p_entities.FeedInfo, is_linked: bool) -> "FeedInfo":
        return cls(
            url=feed.url,
            title=feed.title,
            description=feed.description,
            isLinked=is_linked,
            entries=[EntryInfo.from_internal(entry) for entry in feed.entries],
        )


class TagInfo(BaseEntity):
    uid: str
    name: str
    link: str | None
    categories: set[o_entities.TagCategory]

    @classmethod
    def from_internal(cls, tag: o_entities.Tag, uid: str) -> "TagInfo":
        assert tag.name is not None

        return cls(uid=uid, name=tag.name, link=tag.link, categories=tag.categories)


class UserSettingKind(enum.StrEnum):
    openai_api_key = "openai_api_key"
    gemini_api_key = "gemini_api_key"

    hide_message_about_setting_up_key = "hide_message_about_setting_up_key"
    hide_message_about_adding_collections = "hide_message_about_adding_collections"
    hide_message_check_your_feed_urls = "hide_message_check_your_feed_urls"
    process_entries_not_older_than = "process_entries_not_older_than"
    max_tokens_cost_in_month = "max_tokens_cost_in_month"

    @classmethod
    def from_internal(cls, kind: int) -> "UserSettingKind":
        from ffun.application.user_settings import UserSetting

        real_kind = UserSetting(kind)
        return UserSettingKind(real_kind.name)

    def to_internal(self) -> int:
        from ffun.application.user_settings import UserSetting

        return getattr(UserSetting, self.name)  # type: ignore


class UserSetting(BaseEntity):
    kind: UserSettingKind
    type: us_types.TypeId  # should not differ between front & back => no need to convert
    value: Any
    name: str

    @classmethod
    def from_internal(cls, kind: int, value: str | int | float | bool) -> "UserSetting":
        from ffun.application.user_settings import UserSetting

        real_kind = UserSetting(kind)

        real_setting = user_settings.get(real_kind)

        assert real_setting is not None

        return cls(
            kind=UserSettingKind.from_internal(real_kind),
            type=real_setting.type.id,
            value=real_setting.type.normalize(value),
            name=real_setting.name,
        )


class ResourceKind(enum.StrEnum):
    tokens_cost = "tokens_cost"

    @classmethod
    def from_internal(cls, kind: int) -> "ResourceKind":
        from ffun.application.resources import Resource

        real_kind = Resource(kind)
        return ResourceKind(real_kind.name)

    def to_internal(self) -> int:
        from ffun.application.resources import Resource

        return getattr(Resource, self.name)  # type: ignore


class ResourceHistoryRecord(pydantic.BaseModel):
    intervalStartedAt: datetime.datetime
    used: Decimal
    reserved: Decimal

    @classmethod
    def from_internal(cls, record: r_entities.Resource) -> "ResourceHistoryRecord":
        from ffun.application.resources import Resource

        if record.kind == Resource.tokens_cost:
            transformer = _cost_points.to_cost
        else:

            def transformer(points: int) -> USDCost:
                return USDCost(Decimal(points))

        return cls(
            intervalStartedAt=record.interval_started_at,
            used=transformer(record.used),
            reserved=transformer(record.reserved),
        )


class Collection(pydantic.BaseModel):
    id: CollectionId
    slug: CollectionSlug
    guiOrder: int
    name: str
    description: str
    feedsNumber: int
    showOnMain: bool

    @classmethod
    def from_internal(cls, record: fc_entities.Collection) -> "Collection":
        return cls(
            id=record.id,
            slug=record.slug,
            guiOrder=record.gui_order,
            name=record.name,
            description=record.description,
            feedsNumber=len(record.feeds),
            showOnMain=record.show_on_main,
        )


class CollectionFeedInfo(pydantic.BaseModel):
    url: FeedUrl
    title: str
    description: str
    id: f_entities.FeedId

    @classmethod
    def from_internal(cls, record: fc_entities.FeedInfo) -> "CollectionFeedInfo":
        assert record.feed_id is not None
        return cls(url=record.url, title=record.title, description=record.description, id=record.feed_id)


##################
# Request/Response
##################


class GetFeedsRequest(api.APIRequest):
    pass


class GetFeedsResponse(api.APISuccess):
    feeds: list[Feed]


class GetLastEntriesRequest(api.APIRequest):
    period: datetime.timedelta | None = None
    minTagCount: int = 2  # TODO: remove default after 1.21 branch is released

    @pydantic.field_validator("period")
    def validate_period(cls, v: None | datetime.timedelta) -> None | datetime.timedelta:
        if v is not None and v.total_seconds() < 0:
            raise ValueError("period must be positive")
        return v


class GetLastEntriesResponse(api.APISuccess):
    entries: list[Entry]
    tagsMapping: dict[TagId, TagUid]


class GetLastCollectionEntriesRequest(api.APIRequest):
    collectionSlug: CollectionSlug
    period: datetime.timedelta | None = None
    minTagCount: int = 2  # TODO: remove default after 1.21 branch is released

    @pydantic.field_validator("period")
    def validate_period(cls, v: None | datetime.timedelta) -> None | datetime.timedelta:
        if v is not None and v.total_seconds() < 0:
            raise ValueError("period must be positive")
        return v


class GetLastCollectionEntriesResponse(api.APISuccess):
    entries: list[Entry]
    tagsMapping: dict[TagId, TagUid]


class GetEntriesByIdsRequest(api.APIRequest):
    ids: list[EntryId]
    includeTags: bool = False  # TODO: remove default after 1.21 branch is released


class GetEntriesByIdsResponse(api.APISuccess):
    entries: list[Entry]
    tagsMapping: dict[TagId, TagUid]


class CreateOrUpdateRuleRequest(api.APIRequest):
    requiredTags: list[TagUid]
    excludedTags: list[TagUid]
    score: int


class CreateOrUpdateRuleResponse(api.APISuccess):
    pass


class DeleteRuleRequest(api.APIRequest):
    id: RuleId


class DeleteRuleResponse(api.APISuccess):
    pass


class UpdateRuleRequest(api.APIRequest):
    id: RuleId
    requiredTags: list[TagUid]
    excludedTags: list[TagUid]
    score: int


class UpdateRuleResponse(api.APISuccess):
    pass


class GetRulesRequest(api.APIRequest):
    pass


class GetRulesResponse(api.APISuccess):
    rules: list[Rule]


class GetScoreDetailsRequest(api.APIRequest):
    entryId: EntryId


class GetScoreDetailsResponse(api.APISuccess):
    rules: list[Rule]


class SetMarkerRequest(api.APIRequest):
    entryId: EntryId
    marker: Marker


class SetMarkerResponse(api.APISuccess):
    pass


class RemoveMarkerRequest(api.APIRequest):
    entryId: EntryId
    marker: Marker


class RemoveMarkerResponse(api.APISuccess):
    pass


class DiscoverFeedsRequest(api.APIRequest):
    url: UnknownUrl


class DiscoverFeedsResponse(api.APISuccess):
    feeds: list[FeedInfo]


class AddFeedRequest(api.APIRequest):
    url: UnknownUrl


class AddFeedResponse(api.APISuccess):
    feed: Feed


class AddOpmlRequest(api.APIRequest):
    content: str


class AddOpmlResponse(api.APISuccess):
    pass


class UnsubscribeRequest(api.APIRequest):
    feedId: f_entities.FeedId


class UnsubscribeResponse(api.APISuccess):
    pass


class GetFeedsCollectionsRequest(api.APIRequest):
    pass


class GetFeedsCollectionsResponse(api.APISuccess):
    collections: list[Collection]


class GetCollectionFeedsRequest(api.APIRequest):
    collectionId: fc_entities.CollectionId


class GetCollectionFeedsResponse(api.APISuccess):
    feeds: list[CollectionFeedInfo]


class SubscribeToCollectionsRequest(api.APIRequest):
    collections: list[fc_entities.CollectionId]


class SubscribeToCollectionsResponse(api.APISuccess):
    pass


class GetTagsInfoRequest(api.APIRequest):
    uids: set[TagUid]


class GetTagsInfoResponse(api.APISuccess):
    tags: dict[TagUid, TagInfo]


class GetUserSettingsRequest(api.APIRequest):
    pass


class GetUserSettingsResponse(api.APISuccess):
    settings: list[UserSetting]


class SetUserSettingRequest(api.APIRequest):
    kind: UserSettingKind
    value: Any

    @pydantic.model_validator(mode="before")
    @classmethod
    def validate_value(cls, values: dict[str, Any]) -> dict[str, Any]:
        from ffun.application.user_settings import UserSetting

        kind = UserSettingKind(values["kind"]).to_internal()
        value = values.get("value")

        real_kind = UserSetting(kind)

        real_setting = user_settings.get(real_kind)

        values["value"] = real_setting.type.normalize(value)

        return values


class SetUserSettingResponse(api.APISuccess):
    pass


class GetResourceHistoryRequest(api.APIRequest):
    kind: ResourceKind


class GetResourceHistoryResponse(api.APISuccess):
    history: list[ResourceHistoryRecord]


class GetInfoRequest(api.APIRequest):
    pass


class GetInfoResponse(api.APISuccess):
    userId: UserId


class TrackEventRequest(api.APIRequest):
    event: front_events.Event


class TrackEventResponse(api.APISuccess):
    pass


class RemoveUserRequest(api.APIRequest):
    pass


class RemoveUserResponse(api.APISuccess):
    pass



import pytest

from ffun.core.postgresql import transaction
from ffun.core.tests.helpers import TableSizeDelta, TableSizeNotChanged, assert_logs_has_business_event, capture_logs
from ffun.domain.entities import UserId
from ffun.feeds.entities import FeedId
from ffun.feeds.tests import make as f_make
from ffun.feeds_collections.collections import collections
from ffun.feeds_collections.entities import CollectionId
from ffun.feeds_links.entities import FeedLink
from ffun.feeds_links.operations import (
    add_link,
    count_feeds_per_user,
    count_subset_feeds_per_user,
    get_link,
    get_linked_feeds,
    get_linked_users,
    has_linked_users,
    remove_link,
    tech_merge_feeds,
    tech_remove_all_links,
)
from ffun.users.tests import make as u_make


class TestAddLink:

    @pytest.mark.parametrize("in_collection", [True, False])
    @pytest.mark.asyncio
    async def test_add_link(
        self,
        internal_user_id: UserId,
        saved_feed_id: FeedId,
        in_collection: bool,
        collection_id_for_test_feeds: CollectionId,
    ) -> None:
        if in_collection:
            await collections.add_test_feed_to_collections(collection_id_for_test_feeds, saved_feed_id)

        with capture_logs() as logs:
            async with TableSizeDelta("fl_links", delta=1):
                await add_link(internal_user_id, saved_feed_id)

        assert_logs_has_business_event(
            logs, "feed_linked", user_id=internal_user_id, feed_id=str(saved_feed_id), in_collection=in_collection
        )

        links = await get_linked_feeds(internal_user_id)

        assert links == [FeedLink(user_id=internal_user_id, feed_id=saved_feed_id, created_at=links[0].created_at)]

    @pytest.mark.asyncio
    async def test_duplicated(self, internal_user_id: UserId, saved_feed_id: FeedId) -> None:
        await add_link(internal_user_id, saved_feed_id)

        async with TableSizeNotChanged("fl_links"):
            await add_link(internal_user_id, saved_feed_id)

        links = await get_linked_feeds(internal_user_id)

        assert links == [FeedLink(user_id=internal_user_id, feed_id=saved_feed_id, created_at=links[0].created_at)]

    @pytest.mark.asyncio
    async def test_multiple_links(
        self, five_internal_user_ids: list[UserId], five_saved_feed_ids: list[FeedId]
    ) -> None:
        user_1_id, user_2_id, user_3_id = five_internal_user_ids[:3]

        f = five_saved_feed_ids

        async with TableSizeDelta("fl_links", delta=6):
            await add_link(user_1_id, f[0])
            await add_link(user_1_id, f[1])

            await add_link(user_2_id, f[1])
            await add_link(user_2_id, f[2])
            await add_link(user_2_id, f[3])

            await add_link(user_3_id, f[3])

        links_1 = await get_linked_feeds(user_1_id)

        assert links_1 == [
            FeedLink(user_id=user_1_id, feed_id=f[0], created_at=links_1[0].created_at),
            FeedLink(user_id=user_1_id, feed_id=f[1], created_at=links_1[1].created_at),
        ]

        links_2 = await get_linked_feeds(user_2_id)

        assert links_2 == [
            FeedLink(user_id=user_2_id, feed_id=f[1], created_at=links_2[0].created_at),
            FeedLink(user_id=user_2_id, feed_id=f[2], created_at=links_2[1].created_at),
            FeedLink(user_id=user_2_id, feed_id=f[3], created_at=links_2[2].created_at),
        ]

        links_3 = await get_linked_feeds(user_3_id)

        assert links_3 == [FeedLink(user_id=user_3_id, feed_id=f[3], created_at=links_3[0].created_at)]


class TestRemoveLink:

    @pytest.mark.parametrize("in_collection", [True, False])
    @pytest.mark.asyncio
    async def test_remove_link(
        self,
        internal_user_id: UserId,
        saved_feed_id: FeedId,
        in_collection: bool,
        collection_id_for_test_feeds: CollectionId,
    ) -> None:
        if in_collection:
            await collections.add_test_feed_to_collections(collection_id_for_test_feeds, saved_feed_id)

        await add_link(internal_user_id, saved_feed_id)

        with capture_logs() as logs:
            async with TableSizeDelta("fl_links", delta=-1):
                await remove_link(internal_user_id, saved_feed_id)

        assert_logs_has_business_event(
            logs, "feed_unlinked", user_id=internal_user_id, feed_id=str(saved_feed_id), in_collection=in_collection
        )

        links = await get_linked_feeds(internal_user_id)

        assert links == []

    @pytest.mark.asyncio
    async def test_create_after_remove(self, internal_user_id: UserId, saved_feed_id: FeedId) -> None:
        await add_link(internal_user_id, saved_feed_id)
        await remove_link(internal_user_id, saved_feed_id)

        async with TableSizeDelta("fl_links", delta=1):
            await add_link(internal_user_id, saved_feed_id)

        links = await get_linked_feeds(internal_user_id)

        assert links == [FeedLink(user_id=internal_user_id, feed_id=saved_feed_id, created_at=links[0].created_at)]

    @pytest.mark.asyncio
    async def test_remove_only_required(
        self, five_internal_user_ids: list[UserId], five_saved_feed_ids: list[FeedId]
    ) -> None:
        user_1_id, user_2_id, user_3_id = five_internal_user_ids[:3]

        f = five_saved_feed_ids

        await add_link(user_1_id, f[0])
        await add_link(user_1_id, f[1])

        await add_link(user_2_id, f[1])
        await add_link(user_2_id, f[2])
        await add_link(user_2_id, f[3])

        await add_link(user_3_id, f[3])

        async with TableSizeDelta("fl_links", delta=-1):
            await remove_link(user_2_id, f[2])

        links_1 = await get_linked_feeds(user_1_id)
        assert len(links_1) == 2

        links_2 = await get_linked_feeds(user_2_id)

        assert links_2 == [
            FeedLink(user_id=user_2_id, feed_id=f[1], created_at=links_2[0].created_at),
            FeedLink(user_id=user_2_id, feed_id=f[3], created_at=links_2[1].created_at),
        ]

        links_3 = await get_linked_feeds(user_3_id)

        assert len(links_3) == 1


class TestTechRemoveAllLinks:

    @pytest.mark.asyncio
    async def test_nothing_to_remove(self) -> None:
        async with TableSizeNotChanged("fl_links"):
            await tech_remove_all_links([])

    @pytest.mark.asyncio
    async def test_nothing_no_links_to_remove(self, saved_feed_id: FeedId) -> None:
        async with TableSizeNotChanged("fl_links"):
            await tech_remove_all_links([saved_feed_id])

    @pytest.mark.asyncio
    async def test_remove(self, five_internal_user_ids: list[UserId], five_saved_feed_ids: list[FeedId]) -> None:
        user_1_id, user_2_id, user_3_id = five_internal_user_ids[:3]

        f = five_saved_feed_ids

        await add_link(user_1_id, f[0])
        await add_link(user_1_id, f[1])

        await add_link(user_2_id, f[1])
        await add_link(user_2_id, f[2])
        await add_link(user_2_id, f[3])

        await add_link(user_3_id, f[3])

        async with TableSizeDelta("fl_links", delta=-4):
            await tech_remove_all_links([f[1], f[3]])

        links_1 = await get_linked_feeds(user_1_id)
        assert links_1 == [FeedLink(user_id=user_1_id, feed_id=f[0], created_at=links_1[0].created_at)]

        links_2 = await get_linked_feeds(user_2_id)
        assert links_2 == [FeedLink(user_id=user_2_id, feed_id=f[2], created_at=links_2[0].created_at)]

        links_3 = await get_linked_feeds(user_3_id)
        assert links_3 == []


class TestGetLinkedFeeds:
    """This functions is validated by other tests."""

    @pytest.mark.asyncio
    async def test_empty(self, internal_user_id: UserId) -> None:
        links = await get_linked_feeds(internal_user_id)

        assert links == []


class TestGetLink:

    @pytest.mark.asyncio
    async def test_no_link(self, internal_user_id: UserId, saved_feed_id: FeedId) -> None:
        link = await get_link(internal_user_id, saved_feed_id)

        assert link is None

    @pytest.mark.asyncio
    async def test_has_link(self, internal_user_id: UserId, saved_feed_id: FeedId) -> None:
        await add_link(internal_user_id, saved_feed_id)

        link = await get_link(internal_user_id, saved_feed_id)

        assert link is not None

        assert link == FeedLink(user_id=internal_user_id, feed_id=saved_feed_id, created_at=link.created_at)


class TestGetLinkedUsers:

    @pytest.mark.asyncio
    async def test_empty(self, saved_feed_id: FeedId) -> None:
        users = await get_linked_users([saved_feed_id])

        assert users == {}

    @pytest.mark.asyncio
    async def test_returns_only_required(
        self, five_internal_user_ids: list[UserId], saved_feed_id: FeedId, another_saved_feed_id: FeedId
    ) -> None:
        user_1_id, user_2_id, user_3_id = five_internal_user_ids[:3]

        await add_link(user_1_id, saved_feed_id)

        await add_link(user_2_id, saved_feed_id)
        await add_link(user_2_id, another_saved_feed_id)

        await add_link(user_3_id, another_saved_feed_id)

        users = await get_linked_users([saved_feed_id])

        assert users == {saved_feed_id: {user_1_id, user_2_id}}

        users = await get_linked_users([another_saved_feed_id])

        assert users == {another_saved_feed_id: {user_2_id, user_3_id}}

        users = await get_linked_users([saved_feed_id, another_saved_feed_id])

        assert users == {saved_feed_id: {user_1_id, user_2_id}, another_saved_feed_id: {user_2_id, user_3_id}}


class TestHasLinkedUsers:

    @pytest.mark.asyncio
    async def test_no_users(self, saved_feed_id: FeedId) -> None:
        assert not await has_linked_users(saved_feed_id)

    @pytest.mark.asyncio
    async def test_has_users(self, internal_user_id: UserId, saved_feed_id: FeedId) -> None:
        await add_link(internal_user_id, saved_feed_id)

        assert await has_linked_users(saved_feed_id)


class TestCountFeedsPerUser:

    @pytest.mark.asyncio
    async def test(self, five_internal_user_ids: list[UserId], five_saved_feed_ids: list[FeedId]) -> None:
        u = five_internal_user_ids
        f = five_saved_feed_ids

        await add_link(u[0], f[0])
        await add_link(u[0], f[1])
        await add_link(u[0], f[2])
        await add_link(u[1], f[2])
        await add_link(u[1], f[3])
        await add_link(u[2], f[3])

        numbers_after = await count_feeds_per_user()

        assert numbers_after[u[0]] == 3
        assert numbers_after[u[1]] == 2
        assert numbers_after[u[2]] == 1
        assert u[3] not in numbers_after
        assert u[4] not in numbers_after


class TestCountSubsetFeedsPerUser:

    @pytest.mark.asyncio
    async def test(
        self,
        five_internal_user_ids: list[UserId],
        five_saved_feed_ids: list[FeedId],
        collection_id_for_test_feeds: CollectionId,
    ) -> None:
        u = five_internal_user_ids
        f = five_saved_feed_ids

        await add_link(u[0], f[0])
        await add_link(u[0], f[1])
        await add_link(u[0], f[2])
        await add_link(u[1], f[2])
        await add_link(u[1], f[3])
        await add_link(u[2], f[3])

        numbers_after = await count_subset_feeds_per_user([f[0], f[2]])

        assert numbers_after[u[0]] == 2
        assert numbers_after[u[1]] == 1
        assert u[2] not in numbers_after
        assert u[3] not in numbers_after
        assert u[4] not in numbers_after


class TestMergeFeeds:
    @pytest.mark.asyncio
    async def test_nothing_to_merge(self, loaded_feed_id: FeedId, another_loaded_feed_id: FeedId) -> None:
        async with transaction() as trx:
            await tech_merge_feeds(trx, loaded_feed_id, another_loaded_feed_id)

    @pytest.mark.asyncio
    async def test_move_links(self) -> None:
        feed_1, feed_2, feed_3 = await f_make.n_feeds(3)

        user_a, user_b = await u_make.n_users(2)

        await add_link(user_a, feed_1.id)
        await add_link(user_b, feed_2.id)

        async with TableSizeNotChanged("fl_links"):
            async with transaction() as trx:
                await tech_merge_feeds(trx, from_feed_id=feed_2.id, to_feed_id=feed_3.id)

        links_a = await get_linked_feeds(user_a)
        links_b = await get_linked_feeds(user_b)

        assert {link.feed_id for link in links_a} == {feed_1.id}
        assert {link.feed_id for link in links_b} == {feed_3.id}

    @pytest.mark.asyncio
    async def test_rewrite_links(self) -> None:
        feed_1, feed_2, feed_3 = await f_make.n_feeds(3)

        user_a, user_b = await u_make.n_users(2)

        await add_link(user_a, feed_1.id)
        await add_link(user_b, feed_2.id)
        await add_link(user_b, feed_3.id)

        async with TableSizeDelta("fl_links", delta=-1):
            async with transaction() as trx:
                await tech_merge_feeds(trx, from_feed_id=feed_2.id, to_feed_id=feed_3.id)

        links_a = await get_linked_feeds(user_a)
        links_b = await get_linked_feeds(user_b)

        assert {link.feed_id for link in links_a} == {feed_1.id}
        assert {link.feed_id for link in links_b} == {feed_3.id}

    @pytest.mark.asyncio
    async def test_complex_ops(self) -> None:
        feed_1, feed_2, feed_3 = await f_make.n_feeds(3)

        user_a, user_b, user_c = await u_make.n_users(3)

        await add_link(user_a, feed_1.id)

        await add_link(user_b, feed_2.id)
        await add_link(user_b, feed_3.id)

        await add_link(user_c, feed_2.id)

        async with TableSizeDelta("fl_links", delta=-1):
            async with transaction() as trx:
                await tech_merge_feeds(trx, from_feed_id=feed_2.id, to_feed_id=feed_3.id)

        links_a = await get_linked_feeds(user_a)
        links_b = await get_linked_feeds(user_b)
        links_c = await get_linked_feeds(user_c)

        assert {link.feed_id for link in links_a} == {feed_1.id}
        assert {link.feed_id for link in links_b} == {feed_3.id}
        assert {link.feed_id for link in links_c} == {feed_3.id}



import pathlib

import pydantic_settings

from ffun.core.settings import BaseSettings

_root = pathlib.Path(__file__).parent


class Settings(BaseSettings):

    gemini_api_entry_point: str = "https://generativelanguage.googleapis.com/v1beta"
    gemini_api_timeout: float = 20.0

    model_config = pydantic_settings.SettingsConfigDict(env_prefix="FFUN_GOOGLE_")


settings = Settings()



# TODO: refactor these commands and move them to application.py

from ffun.cli.application import app  # noqa: F401
from ffun.cli.commands import configs  # noqa: F401
from ffun.cli.commands import entries  # noqa: F401
from ffun.cli.commands import failed_entries  # noqa: F401
from ffun.cli.commands import supertokens  # noqa: F401
from ffun.cli.commands import workers  # noqa: F401
from ffun.cli.commands import yoyo  # noqa: F401

__all__ = ["app"]



from ffun.core import utils
from ffun.domain.domain import new_rule_id, new_user_id
from ffun.domain.entities import TagId
from ffun.scores import domain, entities


def rule(score: int, required_tags: set[int], excluded_tags: set[int]) -> entities.Rule:
    return entities.Rule(
        id=new_rule_id(),
        user_id=new_user_id(),
        score=score,
        required_tags={TagId(t) for t in required_tags},
        excluded_tags={TagId(t) for t in excluded_tags},
        created_at=utils.now(),
        updated_at=utils.now(),
    )


def get_score_rules(rules: list[entities.Rule], tags: set[int]) -> list[entities.Rule]:
    return domain.get_score_rules(rules, {TagId(t) for t in tags})


def get_score_contributions(rules: list[entities.Rule], tags: set[int]) -> tuple[int, dict[int, int]]:
    score, contributions = domain.get_score_contributions(rules, {TagId(t) for t in tags})
    return score, {int(k): v for k, v in contributions.items()}


class TestGetScoreRules:

    def test_no_rules(self) -> None:
        assert get_score_rules([], set()) == []
        assert get_score_rules([], {1, 2, 3}) == []

    def test_only_required_tags(self) -> None:
        rules = [rule(2, {1, 2, 3}, set()), rule(3, {1, 3}, set()), rule(5, {2}, set())]

        assert get_score_rules(rules, set()) == []
        assert get_score_rules(rules, {1, 2, 3}) == rules
        assert get_score_rules(rules, {1, 2, 3, 4}) == rules
        assert get_score_rules(rules, {1, 3}) == [rules[1]]
        assert get_score_rules(rules, {2, 3}) == [rules[2]]
        assert get_score_rules(rules, {2}) == [rules[2]]

    def test_only_excluded_tags(self) -> None:
        rules = [rule(2, set(), {1, 2, 3}), rule(3, set(), {1, 3}), rule(5, set(), {2})]

        assert get_score_rules(rules, set()) == rules
        assert get_score_rules(rules, {1, 2, 3}) == []
        assert get_score_rules(rules, {1, 2, 3, 4}) == []
        assert get_score_rules(rules, {1, 3}) == [rules[2]]
        assert get_score_rules(rules, {2, 3}) == []
        assert get_score_rules(rules, {2}) == [rules[1]]
        assert get_score_rules(rules, {1}) == [rules[2]]

    def test_required_and_excluded_tags(self) -> None:
        rules = [rule(2, {1, 2, 3}, {4}), rule(3, {1, 3}, {2}), rule(5, {2}, {1, 3})]

        assert get_score_rules(rules, set()) == []
        assert get_score_rules(rules, {1, 2, 3}) == [rules[0]]
        assert get_score_rules(rules, {1, 2, 3, 4}) == []
        assert get_score_rules(rules, {1, 3}) == [rules[1]]
        assert get_score_rules(rules, {2, 3}) == []
        assert get_score_rules(rules, {2}) == [rules[2]]
        assert get_score_rules(rules, {1, 3, 4}) == [rules[1]]


class TestGetScore:

    def test_no_rules(self) -> None:
        assert get_score_contributions([], set()) == (0, {})
        assert get_score_contributions([], {1, 2, 3}) == (0, {})

    def test_only_required_tags(self) -> None:
        rules = [rule(2, {1, 2, 3}, set()), rule(3, {1, 3}, set()), rule(5, {2}, set())]

        assert get_score_contributions(rules, set()) == (0, {})
        assert get_score_contributions(rules, {1, 2, 3}) == (10, {1: 5, 2: 7, 3: 5})
        assert get_score_contributions(rules, {1, 2, 3, 4}) == (10, {1: 5, 2: 7, 3: 5})
        assert get_score_contributions(rules, {1, 3}) == (3, {1: 3, 3: 3})
        assert get_score_contributions(rules, {2, 3}) == (5, {2: 5})
        assert get_score_contributions(rules, {2}) == (5, {2: 5})

    def test_only_excluded_tags(self) -> None:
        rules = [rule(2, set(), {1, 2, 3}), rule(3, set(), {1, 3}), rule(5, set(), {2})]

        assert get_score_contributions(rules, set()) == (10, {})
        assert get_score_contributions(rules, {1, 2, 3}) == (0, {})
        assert get_score_contributions(rules, {1, 2, 3, 4}) == (0, {})
        assert get_score_contributions(rules, {1, 3}) == (5, {})
        assert get_score_contributions(rules, {2, 3}) == (0, {})
        assert get_score_contributions(rules, {2}) == (3, {})
        assert get_score_contributions(rules, {1}) == (5, {})

    def test_required_and_excluded_tags(self) -> None:
        rules = [rule(2, {1, 2, 3}, {4}), rule(3, {1, 3}, {2}), rule(5, {2}, {1, 3})]

        assert get_score_contributions(rules, set()) == (0, {})
        assert get_score_contributions(rules, {1, 2, 3}) == (2, {1: 2, 2: 2, 3: 2})
        assert get_score_contributions(rules, {1, 2, 3, 4}) == (0, {})
        assert get_score_contributions(rules, {1, 3}) == (3, {1: 3, 3: 3})
        assert get_score_contributions(rules, {2, 3}) == (0, {})
        assert get_score_contributions(rules, {2}) == (5, {2: 5})
        assert get_score_contributions(rules, {1, 3, 4}) == (3, {1: 3, 3: 3})



import pytest
from structlog.testing import capture_logs

from ffun.core.postgresql import execute
from ffun.core.tests.helpers import assert_logs
from ffun.domain.domain import new_entry_id
from ffun.feeds.entities import Feed
from ffun.feeds_collections.collections import collections
from ffun.feeds_collections.entities import CollectionId
from ffun.librarian import operations
from ffun.librarian.background_processors import EntriesProcessor
from ffun.librarian.tests import make
from ffun.library import domain as l_domain
from ffun.library.tests import make as l_make
from ffun.ontology import domain as o_domain


class TestEntriesProcessors:
    @pytest.mark.asyncio
    async def test_no_entries_to_process(self, fake_entries_processor: EntriesProcessor) -> None:
        await operations.clear_processor_queue(fake_entries_processor.id)
        await make.end_processor_pointer(fake_entries_processor.id)

        with capture_logs() as logs:
            await fake_entries_processor.single_run()

        assert_logs(logs, no_entries_to_process=1)

    @pytest.mark.asyncio
    async def test_entries_more_than_concurrency(
        self, fake_entries_processor: EntriesProcessor, loaded_feed: Feed
    ) -> None:
        await operations.clear_processor_queue(fake_entries_processor.id)
        await make.end_processor_pointer(fake_entries_processor.id)

        entries = await l_make.n_entries(loaded_feed, 9)
        entries_list = list(entries.values())
        entries_list.sort(key=lambda entry: (entry.cataloged_at, entry.id))

        assert fake_entries_processor.concurrency <= len(entries)

        with capture_logs() as logs:
            await fake_entries_processor.single_run()

        assert_logs(logs, no_entries_to_process=0)

        tags = await o_domain.get_tags_ids_for_entries(list(entries))

        expected_ids = await o_domain.get_ids_by_uids({"fake-constant-tag-1", "fake-constant-tag-2"})  # type: ignore

        for entry in entries_list[: fake_entries_processor.concurrency]:
            assert tags[entry.id] == set(expected_ids.values())

        for entry in entries_list[fake_entries_processor.concurrency :]:
            assert entry.id not in tags

    @pytest.mark.asyncio
    async def test_entries_less_than_concurrency(
        self, fake_entries_processor: EntriesProcessor, loaded_feed: Feed
    ) -> None:
        await operations.clear_processor_queue(fake_entries_processor.id)
        await make.end_processor_pointer(fake_entries_processor.id)

        entries = await l_make.n_entries(loaded_feed, 2)
        entries_list = list(entries.values())
        entries_list.sort(key=lambda entry: (entry.cataloged_at, entry.id))

        assert fake_entries_processor.concurrency > len(entries)

        with capture_logs() as logs:
            await fake_entries_processor.single_run()

        assert_logs(logs, no_entries_to_process=0)

        tags = await o_domain.get_tags_ids_for_entries(list(entries))

        expected_ids = await o_domain.get_ids_by_uids({"fake-constant-tag-1", "fake-constant-tag-2"})  # type: ignore

        for entry in entries_list:
            assert tags[entry.id] == set(expected_ids.values())

    @pytest.mark.asyncio
    async def test_unexisted_entries_in_queue(
        self, fake_entries_processor: EntriesProcessor, loaded_feed: Feed
    ) -> None:
        await operations.clear_processor_queue(fake_entries_processor.id)
        await make.end_processor_pointer(fake_entries_processor.id)

        entries = await l_make.n_entries(loaded_feed, 2)
        entries_list = list(entries.values())
        entries_list.sort(key=lambda entry: (entry.cataloged_at, entry.id))

        fake_entries_ids = [new_entry_id() for _ in range(3)]

        await operations.push_entries_to_processor_queue(
            execute, processor_id=fake_entries_processor.id, entry_ids=fake_entries_ids
        )

        assert fake_entries_processor.concurrency >= len(entries) + len(fake_entries_ids)

        with capture_logs() as logs:
            await fake_entries_processor.single_run()

        assert_logs(logs, no_entries_to_process=0, unexisted_entry_in_queue=3)

        tags = await o_domain.get_tags_ids_for_entries(list(entries))

        expected_ids = await o_domain.get_ids_by_uids({"fake-constant-tag-1", "fake-constant-tag-2"})  # type: ignore

        for entry in entries_list:
            assert tags[entry.id] == set(expected_ids.values())

        entities_in_queue = await operations.get_entries_to_process(processor_id=fake_entries_processor.id, limit=100)

        assert entities_in_queue == []

    @pytest.mark.asyncio
    async def test_separate_entries__unexsisted_entries(
        self, fake_entries_processor: EntriesProcessor, loaded_feed: Feed
    ) -> None:
        entries = await l_make.n_entries(loaded_feed, 2)
        entries_list = list(entries.values())
        entries_list.sort(key=lambda entry: (entry.cataloged_at, entry.id))

        fake_entries_ids = [new_entry_id() for _ in range(3)]

        entries_ids = [entry.id for entry in entries_list] + fake_entries_ids

        with capture_logs() as logs:
            entries_to_process, entries_to_remove = await fake_entries_processor.separate_entries(
                entries_ids=entries_ids
            )

        assert_logs(
            logs,
            unexisted_entry_in_queue=3,
            entry_without_feeds_in_queue=0,
            proccessor_not_allowed_for_collections=0,
            proccessor_not_allowed_for_users=0,
            proccessor_is_allowed_for_entry=2,
        )

        assert {entry.id for entry in entries_to_process} == {entry.id for entry in entries_list}
        assert set(entries_to_remove) == set(fake_entries_ids)

    @pytest.mark.asyncio
    async def test_separate_entries__entries_without_feeds_in_queue(
        self, fake_entries_processor: EntriesProcessor, loaded_feed: Feed
    ) -> None:
        entries = await l_make.n_entries(loaded_feed, 5)
        entries_list = list(entries.values())
        entries_list.sort(key=lambda entry: (entry.cataloged_at, entry.id))

        await l_domain.unlink_feed_tail(loaded_feed.id, offset=3)

        entries_ids = [entry.id for entry in entries_list]

        with capture_logs() as logs:
            entries_to_process, entries_to_remove = await fake_entries_processor.separate_entries(
                entries_ids=entries_ids
            )

        assert_logs(
            logs,
            unexisted_entry_in_queue=0,
            entry_without_feeds_in_queue=2,
            proccessor_not_allowed_for_collections=0,
            proccessor_not_allowed_for_users=0,
            proccessor_is_allowed_for_entry=3,
        )

        assert {entry.id for entry in entries_to_process} == {entry.id for entry in entries_list[2:]}
        assert set(entries_to_remove) == {entry.id for entry in entries_list[:2]}

    @pytest.mark.asyncio
    async def test_separate_entries__collections_not_allowed(
        self,
        fake_entries_processor: EntriesProcessor,
        loaded_feed: Feed,
        another_loaded_feed: Feed,
        collection_id_for_test_feeds: CollectionId,
    ) -> None:
        entries_1 = await l_make.n_entries(loaded_feed, 3)
        entries_2 = await l_make.n_entries(another_loaded_feed, 2)

        await collections.add_test_feed_to_collections(collection_id_for_test_feeds, another_loaded_feed.id)

        entries_list = list(entries_1.values()) + list(entries_2.values())
        entries_list.sort(key=lambda entry: (entry.cataloged_at, entry.id))

        entries_ids = [entry.id for entry in entries_list]

        fake_entries_processor._processor_info.allowed_for_collections = False

        with capture_logs() as logs:
            entries_to_process, entries_to_remove = await fake_entries_processor.separate_entries(
                entries_ids=entries_ids
            )

        assert_logs(
            logs,
            unexisted_entry_in_queue=0,
            proccessor_not_allowed_for_collections=2,
            proccessor_not_allowed_for_users=0,
            proccessor_is_allowed_for_entry=3,
        )

        assert {entry.id for entry in entries_to_process} == set(entries_1)
        assert set(entries_to_remove) == set(entries_2)

    @pytest.mark.asyncio
    async def test_separate_entries__collections_not_allowed__entry_is_in_multiple_feeds(
        self,
        fake_entries_processor: EntriesProcessor,
        loaded_feed: Feed,
        another_loaded_feed: Feed,
        collection_id_for_test_feeds: CollectionId,
    ) -> None:
        entries_1 = await l_make.n_entries_list(loaded_feed, 3)
        entries_2 = await l_make.n_entries_list(another_loaded_feed, 2)

        await l_domain.catalog_entries(another_loaded_feed.id, entries_1[:2])

        await collections.add_test_feed_to_collections(collection_id_for_test_feeds, another_loaded_feed.id)

        entries_list = list(entries_1) + list(entries_2)
        entries_list.sort(key=lambda entry: (entry.cataloged_at, entry.id))

        entries_ids = [entry.id for entry in entries_list]

        fake_entries_processor._processor_info.allowed_for_collections = False

        with capture_logs() as logs:
            entries_to_process, entries_to_remove = await fake_entries_processor.separate_entries(
                entries_ids=entries_ids
            )

        assert_logs(
            logs,
            unexisted_entry_in_queue=0,
            proccessor_not_allowed_for_collections=4,
            proccessor_not_allowed_for_users=0,
            proccessor_is_allowed_for_entry=1,
        )

        assert {entry.id for entry in entries_to_process} == {entries_1[2].id}
        assert set(entries_to_remove) == {entry.id for entry in entries_1[:2]} | {entry.id for entry in entries_2}

    @pytest.mark.asyncio
    async def test_separate_entries__all_allowed(
        self,
        fake_entries_processor: EntriesProcessor,
        loaded_feed: Feed,
        another_loaded_feed: Feed,
        collection_id_for_test_feeds: CollectionId,
    ) -> None:
        entries_1 = await l_make.n_entries(loaded_feed, 3)
        entries_2 = await l_make.n_entries(another_loaded_feed, 2)

        await collections.add_test_feed_to_collections(collection_id_for_test_feeds, another_loaded_feed.id)

        entries_list = list(entries_1.values()) + list(entries_2.values())
        entries_list.sort(key=lambda entry: (entry.cataloged_at, entry.id))

        entries_ids = [entry.id for entry in entries_list]

        fake_entries_processor._processor_info.allowed_for_collections = True
        fake_entries_processor._processor_info.allowed_for_users = True

        with capture_logs() as logs:
            entries_to_process, entries_to_remove = await fake_entries_processor.separate_entries(
                entries_ids=entries_ids
            )

        assert_logs(
            logs,
            unexisted_entry_in_queue=0,
            proccessor_not_allowed_for_collections=0,
            proccessor_not_allowed_for_users=0,
            proccessor_is_allowed_for_entry=5,
        )

        assert {entry.id for entry in entries_to_process} == set(entries_1) | set(entries_2)

    @pytest.mark.asyncio
    async def test_separate_entries__users_not_allowed(
        self,
        fake_entries_processor: EntriesProcessor,
        loaded_feed: Feed,
        another_loaded_feed: Feed,
        collection_id_for_test_feeds: CollectionId,
    ) -> None:
        entries_1 = await l_make.n_entries(loaded_feed, 3)
        entries_2 = await l_make.n_entries(another_loaded_feed, 2)

        await collections.add_test_feed_to_collections(collection_id_for_test_feeds, another_loaded_feed.id)

        entries_list = list(entries_1.values()) + list(entries_2.values())
        entries_list.sort(key=lambda entry: (entry.cataloged_at, entry.id))

        entries_ids = [entry.id for entry in entries_list]

        fake_entries_processor._processor_info.allowed_for_users = False

        with capture_logs() as logs:
            entries_to_process, entries_to_remove = await fake_entries_processor.separate_entries(
                entries_ids=entries_ids
            )

        assert_logs(
            logs,
            unexisted_entry_in_queue=0,
            proccessor_not_allowed_for_collections=0,
            proccessor_not_allowed_for_users=3,
            proccessor_is_allowed_for_entry=2,
        )

        assert {entry.id for entry in entries_to_process} == set(entries_2)
        assert set(entries_to_remove) == set(entries_1)



import pydantic
import pydantic_settings

from ffun.core.settings import BaseSettings


class PostgreSQL(pydantic.BaseModel):
    host: str = "postgresql"
    port: int = 5432
    user: str = "ffun"
    password: str = "ffun"
    database: str = "ffun"

    pool_min_size: int = 20
    pool_max_size: int | None = None
    pool_timeout: float = 1.0
    pool_num_workers: int = 1
    pool_max_lifetime: int = 9 * 60
    pool_check_period: int = 60

    @property
    def dsn(self) -> str:
        return f"postgresql://{self.user}:{self.password}@{self.host}:{self.port}/{self.database}"

    @property
    def dsn_yoyo(self) -> str:
        return f"postgresql+psycopg://{self.user}:{self.password}@{self.host}:{self.port}/{self.database}"


class Sentry(pydantic.BaseModel):
    dsn: str = ""
    sample_rate: float = 1.0


_development_origins = ("*",)


class Settings(BaseSettings):
    app_name: str = "Feeds Fun"
    app_domain: str = "localhost"
    app_port: int = 5173

    environment: str = "local"

    api_port: int = 5174

    enable_api: bool = False
    enable_sentry: bool = False

    postgresql: PostgreSQL = PostgreSQL()
    sentry: Sentry = Sentry()

    origins: tuple[str, ...] = _development_origins

    @pydantic.model_validator(mode="after")
    def origin_must_be_redefined_in_prod(self) -> "Settings":
        if self.environment == "prod" and self.origins == _development_origins:
            raise ValueError("Origins must be redefined in prod")

        return self

    model_config = pydantic_settings.SettingsConfigDict(env_prefix="FFUN_")


settings = Settings()



"""
feeds-to-entries-many-to-many
"""

from typing import Any

from psycopg import Connection
from psycopg.rows import dict_row
from yoyo import step

__depends__ = {"20240928_02_lD7Bv-source-for-entries"}

sql_create_feeds_to_entries_table = """
CREATE TABLE l_feeds_to_entries (
    feed_id UUID NOT NULL,
    entry_id UUID NOT NULL,
    created_at TIMESTAMP WITH TIME ZONE DEFAULT NOW(),

    PRIMARY KEY (feed_id, entry_id),
    FOREIGN KEY (entry_id) REFERENCES l_entries (id)
)
"""


sql_create_orphaned_entries_table = """
CREATE TABLE l_orphaned_entries (
    entry_id UUID PRIMARY KEY,
    created_at TIMESTAMP WITH TIME ZONE DEFAULT NOW()
)
"""


sql_fill_from_entries_table = """
INSERT INTO l_feeds_to_entries (feed_id, entry_id, created_at)
SELECT DISTINCT ON (source_id, external_id) feed_id, id, created_at
FROM l_entries
ORDER BY source_id, external_id ASC
"""


sql_fill_orphaned_entries = """
INSERT INTO l_orphaned_entries (entry_id)
SELECT id FROM l_entries
LEFT JOIN l_feeds_to_entries ON l_entries.id = l_feeds_to_entries.entry_id
WHERE l_feeds_to_entries.entry_id IS NULL
"""


sql_fill_duplicated_entries = """
WITH duplicates AS (
  SELECT id, source_id, external_id, l_entries.feed_id, l_entries.created_at
  FROM l_entries
  RIGHT JOIN l_orphaned_entries ON l_entries.id = l_orphaned_entries.entry_id
),

originals AS (
  SELECT id, source_id, external_id
  FROM l_entries
  LEFT JOIN l_feeds_to_entries ON l_entries.id = l_feeds_to_entries.entry_id
  WHERE l_feeds_to_entries.entry_id IS NOT NULL
)

INSERT INTO l_feeds_to_entries (feed_id, entry_id, created_at)
SELECT d.feed_id, o.id, d.created_at
FROM duplicates AS d
LEFT JOIN originals AS o ON d.source_id = o.source_id AND d.external_id = o.external_id
"""


sql_remove_duplicated_entries = """
DELETE FROM l_entries AS e
USING l_orphaned_entries AS o
WHERE e.id = o.entry_id
"""


# rollback migration
sql_update_entries_from_m2m_table = """
UPDATE l_entries
SET feed_id = x.feed_id
FROM (SELECT DISTINCT ON (entry_id) feed_id, entry_id FROM l_feeds_to_entries ORDER BY entry_id, created_at ASC) AS x
WHERE x.entry_id = l_entries.id
"""


def apply_step(conn: Connection[dict[str, Any]]) -> None:
    cursor = conn.cursor(row_factory=dict_row)

    cursor.execute(sql_create_feeds_to_entries_table)

    cursor.execute(sql_create_orphaned_entries_table)

    cursor.execute(sql_fill_from_entries_table)

    cursor.execute(sql_fill_orphaned_entries)

    cursor.execute("CREATE INDEX l_feeds_to_entries_entry_id_idx ON l_feeds_to_entries (entry_id)")

    cursor.execute(sql_fill_duplicated_entries)

    cursor.execute(sql_remove_duplicated_entries)

    # external_id goes first to optimize queries
    cursor.execute("CREATE UNIQUE INDEX l_entries_source_id_external_id_idx ON l_entries (external_id, source_id)")

    cursor.execute("ALTER TABLE l_entries DROP COLUMN feed_id")

    cursor.execute(
        """CREATE INDEX l_feeds_to_entries_feed_id_created_at_entity_id_idx
           ON l_feeds_to_entries(feed_id, created_at DESC, entry_id)"""
    )

    cursor.execute("CREATE INDEX l_entries_created_at_idx ON l_entries (created_at ASC)")


def rollback_step(conn: Connection[dict[str, Any]]) -> None:
    cursor = conn.cursor()

    cursor.execute("DROP INDEX l_entries_source_id_external_id_idx")

    cursor.execute("ALTER TABLE l_entries ADD COLUMN feed_id UUID DEFAULT NULL")

    cursor.execute(sql_update_entries_from_m2m_table)

    cursor.execute("ALTER TABLE l_entries ALTER COLUMN feed_id DROP DEFAULT")

    # after some tests there may be entries without a feed
    cursor.execute("DELETE FROM l_entries WHERE feed_id IS NULL")

    cursor.execute("ALTER TABLE l_entries ALTER COLUMN feed_id SET NOT NULL")

    cursor.execute("DROP TABLE l_orphaned_entries")
    cursor.execute("DROP TABLE l_feeds_to_entries")

    cursor.execute("CREATE UNIQUE INDEX l_entries_feed_id_external_id_key ON l_entries (feed_id, external_id)")
    cursor.execute("CREATE INDEX idx_l_entries_created_at_feed_id ON l_entries (created_at, feed_id)")


steps = [step(apply_step, rollback_step)]



from ffun.librarian.processors import base
from ffun.library.entities import Entry
from ffun.ontology.entities import ProcessorTag, TagCategory


class Processor(base.Processor):
    __slots__ = ()

    async def process(self, entry: Entry) -> list[ProcessorTag]:
        tags: list[ProcessorTag] = []

        for external_tag in entry.external_tags:
            tags.append(ProcessorTag(raw_uid=external_tag, categories={TagCategory.feed_tag}))

        return tags



import asyncio
import contextlib
from typing import AsyncGenerator

from ffun.core import logging
from ffun.librarian.background_processors import create_background_processors
from ffun.loader.background_loader import FeedsLoader

logger = logging.get_module_logger()


@contextlib.asynccontextmanager
async def use_loader() -> AsyncGenerator[None, None]:
    logger.info("feeds_loader_enabled")

    feeds_loader = FeedsLoader(name="ffun_feeds_loader", delay_between_runs=1)

    feeds_loader.start()

    logger.info("feeds_loader_initialized")

    try:
        yield
    finally:
        logger.info("deinitialize_feeds_loader")
        await feeds_loader.stop()
        logger.info("feeds_loader_deinitialized")


@contextlib.asynccontextmanager
async def use_librarian() -> AsyncGenerator[None, None]:
    logger.info("librarian_enabled")

    entries_processors = create_background_processors()

    for processor in entries_processors:
        processor.start()

    logger.info("librarian_initialized")

    try:
        yield
    finally:
        logger.info("deinitialize_librarian")
        await asyncio.gather(*[processor.stop() for processor in entries_processors], return_exceptions=True)
        logger.info("librarian_deinitialized")



import asyncio

import typer

from ffun.application.application import with_app
from ffun.core import logging
from ffun.meta.domain import clean_orphaned_entries, clean_orphaned_feeds

logger = logging.get_module_logger()

cli_app = typer.Typer()


async def run_clean(chunk: int) -> None:
    async with with_app():
        logger.info("cleaning_started", chunk=chunk)

        logger.info("cleaning_orphaned_entries_started", chunk=chunk)

        while await clean_orphaned_entries(chunk=chunk) != 0:
            pass

        logger.info("cleaning_orphaned_entries_finished")

        logger.info("cleaning_orphaned_feeds_started", chunk=chunk)

        while await clean_orphaned_feeds(chunk=chunk) != 0:
            pass

        logger.info("cleaning_orphaned_feeds_finished")

        logger.info("cleaning_finished")


@cli_app.command()
def clean(chunk: int = 10000) -> None:
    asyncio.run(run_clean(chunk=chunk))



from decimal import Decimal

import pytest

from ffun.application.resources import Resource as AppResource
from ffun.domain.datetime_intervals import month_interval_start
from ffun.domain.entities import UserId
from ffun.library.entities import Entry
from ffun.llms_framework import errors
from ffun.llms_framework.domain import call_llm, search_for_api_key, split_text, split_text_according_to_tokens
from ffun.llms_framework.entities import APIKeyUsage, LLMApiKey, LLMConfiguration, LLMGeneralApiKey, LLMTokens, USDCost
from ffun.llms_framework.keys_rotator import _cost_points
from ffun.llms_framework.provider_interface import ChatRequestTest, ChatResponseTest, ProviderTest
from ffun.resources import domain as r_domain


class TestSplitText:

    @pytest.mark.parametrize("parts_number", [-100, -1, 0])
    def test_wrong_parts_number(self, parts_number: int) -> None:
        with pytest.raises(errors.TextPartsMustBePositive):
            split_text("some-text", parts=parts_number, intersection=0)

    @pytest.mark.parametrize("intersection", [-100, -1])
    def test_intersection_size(self, intersection: int) -> None:
        with pytest.raises(errors.TextIntersectionMustBePositiveOrZero):
            split_text("some-text", parts=1, intersection=intersection)

    def test_text_is_empty(self) -> None:
        with pytest.raises(errors.TextIsEmpty):
            split_text("", parts=1, intersection=0)

    def test_text_is_too_short(self) -> None:
        with pytest.raises(errors.TextIsTooShort):
            split_text("short", parts=len("short") + 1, intersection=0)

        with pytest.raises(errors.TextIsTooShort):
            split_text("some-text", parts=4, intersection=0)

    @pytest.mark.parametrize("text", ["small-text", "long-long-text " * 10**6], ids=["small", "big"])
    def test_single_part(self, text: str) -> None:
        for intersection in [0, 1, 100, 1000, 10000]:
            assert split_text(text, parts=1, intersection=intersection) == [text]

    @pytest.mark.parametrize(
        "text, parts, intersection, expected",
        [
            ("some-text", 1, 0, ["some-text"]),
            ("some-text", 2, 0, ["some-", "text"]),
            ("some-text", 3, 0, ["som", "e-t", "ext"]),
            ("some-text", 1, 1, ["some-text"]),
            ("some-text", 2, 1, ["some-t", "-text"]),
            ("some-text", 3, 1, ["some", "me-te", "text"]),
            ("some-text", 1, 2, ["some-text"]),
            ("some-text", 2, 2, ["some-te", "e-text"]),
            ("some-text", 3, 2, ["some-", "ome-tex", "-text"]),
            ("some-text", 1, 3, ["some-text"]),
            ("some-text", 2, 3, ["some-tex", "me-text"]),
            ("some-text", 3, 3, ["some-t", "some-text", "e-text"]),
        ],
    )
    def test_split(self, text: str, parts: int, intersection: int, expected: list[str]) -> None:
        assert split_text(text, parts=parts, intersection=intersection) == expected


class TestSplitTextAccordingToTokens:

    @pytest.fixture
    def llm_config(self) -> LLMConfiguration:
        return LLMConfiguration(
            model="test-model-1",
            system="system prompt",
            max_return_tokens=LLMTokens(143),
            text_parts_intersection=100,
            temperature=0,
            top_p=0,
            presence_penalty=0,
            frequency_penalty=0,
        )

    def test_single_part(self, fake_llm_provider: ProviderTest, llm_config: LLMConfiguration) -> None:
        text = "some text"

        parts = split_text_according_to_tokens(llm=fake_llm_provider, llm_config=llm_config, text=text)

        assert parts == [text]

    def test_multiple_parts(self, fake_llm_provider: ProviderTest, llm_config: LLMConfiguration) -> None:
        # We test that algorithm will pass a few iterations before finding the right split
        # => for this test, text must be splittable to 3 parts

        model = fake_llm_provider.get_model(llm_config)

        size = int(model.max_context_size * 2.5)

        text = "a" * size

        parts = split_text_according_to_tokens(llm=fake_llm_provider, llm_config=llm_config, text=text)

        assert len(parts) == 3

        assert size < sum(len(part) for part in parts) < size + 2 * 2 * llm_config.text_parts_intersection + 1

        assert abs(len(parts[0]) - len(parts[1])) <= llm_config.text_parts_intersection + 1
        assert abs(len(parts[1]) - len(parts[2])) <= llm_config.text_parts_intersection + 1
        assert abs(len(parts[0]) - len(parts[2])) <= 1

    def test_trim_if_there_are_too_many_tokens_for_entry(
        self, fake_llm_provider: ProviderTest, llm_config: LLMConfiguration
    ) -> None:
        model = fake_llm_provider.get_model(llm_config)

        text = "a" * (model.max_tokens_per_entry + 1)

        trimmed_parts = split_text_according_to_tokens(llm=fake_llm_provider, llm_config=llm_config, text=text)

        head = "".join(trimmed_parts)

        assert head != text
        assert len("".join(trimmed_parts)) <= len(text)
        assert head == text[: len(head)]


class TestSearchForAPIKey:

    @pytest.fixture
    def llm_config(self) -> LLMConfiguration:
        return LLMConfiguration(
            model="test-model-1",
            system="system prompt",
            max_return_tokens=LLMTokens(143),
            text_parts_intersection=100,
            temperature=0,
            top_p=0,
            presence_penalty=0,
            frequency_penalty=0,
        )

    @pytest.mark.asyncio
    async def test_key_found(
        self,
        fake_llm_api_key: LLMApiKey,
        fake_llm_provider: ProviderTest,
        llm_config: LLMConfiguration,
        cataloged_entry: Entry,
    ) -> None:
        # Here we test meta logic => we can check the simplest case (general api key)
        # plus check the all parameters passed to selection function

        text = "some-text"

        requests = fake_llm_provider.prepare_requests(llm_config, text)

        key_usage = await search_for_api_key(
            llm=fake_llm_provider,
            llm_config=llm_config,
            entry=cataloged_entry,
            requests=requests,
            collections_api_key=None,
            general_api_key=LLMGeneralApiKey(fake_llm_api_key),
        )

        assert key_usage is not None

        assert key_usage.api_key == fake_llm_api_key

    @pytest.mark.asyncio
    async def test_key_not_found(
        self,
        fake_llm_provider: ProviderTest,
        llm_config: LLMConfiguration,
        cataloged_entry: Entry,
    ) -> None:
        # Here we test meta logic => we can check the simplest case (general api key)
        # plus check the all parameters passed to selection function

        text = "some-text"

        requests = fake_llm_provider.prepare_requests(llm_config, text)

        key_usage = await search_for_api_key(
            llm=fake_llm_provider,
            llm_config=llm_config,
            entry=cataloged_entry,
            requests=requests,
            collections_api_key=None,
            general_api_key=None,
        )

        assert key_usage is None


class TestCallLLM:

    @pytest.fixture
    def llm_config(self) -> LLMConfiguration:
        return LLMConfiguration(
            model="test-model-1",
            system="system prompt",
            max_return_tokens=LLMTokens(143),
            text_parts_intersection=100,
            temperature=0,
            top_p=0,
            presence_penalty=0,
            frequency_penalty=0,
        )

    @pytest.mark.asyncio
    async def test_counts_tokens(
        self,
        fake_llm_provider: ProviderTest,
        llm_config: LLMConfiguration,
        internal_user_id: UserId,
        fake_llm_api_key: LLMApiKey,
    ) -> None:

        model = fake_llm_provider.get_model(llm_config)

        interval_started_at = month_interval_start()

        key_usage = APIKeyUsage(
            provider=fake_llm_provider.provider,
            user_id=internal_user_id,
            api_key=fake_llm_api_key,
            reserved_cost=USDCost(Decimal(1005)),
            used_cost=None,
            input_tokens=None,
            output_tokens=None,
            interval_started_at=interval_started_at,
        )

        await r_domain.try_to_reserve(
            user_id=internal_user_id,
            kind=AppResource.tokens_cost,
            interval_started_at=interval_started_at,
            amount=_cost_points.to_points(key_usage.reserved_cost),
            limit=_cost_points.to_points(USDCost(Decimal(124512512))),
        )

        requests = [ChatRequestTest(text="abcd"), ChatRequestTest(text="efgh1234"), ChatRequestTest(text="99")]

        responses = await call_llm(
            llm=fake_llm_provider, llm_config=llm_config, api_key_usage=key_usage, requests=requests
        )

        assert responses == [ChatResponseTest(content=request.text) for request in requests]

        resources = await r_domain.load_resources(
            user_ids=[internal_user_id], kind=AppResource.tokens_cost, interval_started_at=interval_started_at
        )

        cost = USDCost(
            Decimal(
                sum(
                    model.tokens_cost(
                        input_tokens=LLMTokens(len(request.text)), output_tokens=LLMTokens(len(request.text))
                    )
                    for request in requests
                )
            )
        )

        assert resources[internal_user_id].used == _cost_points.to_points(cost)



import pathlib

import toml

from ffun.core import logging
from ffun.domain.domain import new_feed_id
from ffun.domain.entities import CollectionId, CollectionSlug
from ffun.domain.urls import url_to_source_uid
from ffun.feeds import domain as f_domain
from ffun.feeds.domain import get_source_ids
from ffun.feeds.entities import Feed, FeedId
from ffun.feeds_collections import errors
from ffun.feeds_collections.entities import Collection, FeedInfo
from ffun.feeds_collections.settings import settings

logger = logging.get_module_logger()


class Collections:
    __slots__ = ("_collections", "initialized", "_feeds_in_collections")

    def __init__(self) -> None:
        self.initialized = False
        self._collections: list[Collection] = []
        self._feeds_in_collections: dict[FeedId, set[CollectionId]] = {}

    def collections(self) -> list[Collection]:
        return list(self._collections)

    def all_feed_ids(self) -> list[FeedId]:
        return list(self._feeds_in_collections.keys())

    def count_total_feeds(self) -> int:
        return len(self._feeds_in_collections)

    def collection(self, collection_id: CollectionId) -> Collection:
        for collection in self._collections:
            if collection.id == collection_id:
                return collection

        raise errors.CollectionNotFound()

    def collection_by_slug(self, slug: CollectionSlug) -> Collection:
        for collection in self._collections:
            if collection.slug == slug:
                return collection

        raise errors.CollectionNotFound()

    def load(self, collection_configs: pathlib.Path) -> None:
        """Loads all collection configs from the given directory."""
        collections = []

        for path in collection_configs.glob("*.toml"):
            data = toml.loads(path.read_text())

            collection = Collection(**data)

            collections.append(collection)

        self._collections = collections

    def validate_collection_ids(self) -> None:
        ids = {c.id for c in self._collections}

        if len(ids) != len(self._collections):
            raise errors.DuplicateCollectionIds()

    def validate_collection_slugs(self) -> None:
        ids = {c.slug for c in self._collections}

        if len(ids) != len(self._collections):
            raise errors.DuplicateCollectionSlugs()

    def validate_collection_gui_order(self) -> None:
        orders = {c.gui_order for c in self._collections}

        if len(orders) != len(self._collections):
            raise errors.DuplicateCollectionOrders()

    def validate_collection_is_not_empty(self) -> None:
        for collection in self._collections:
            if not collection.feeds:
                raise errors.CollectionIsEmpty()

    # This method may become a bottleneck if there are too many feeds in collections.
    # It could be optimized/removed by refactoring collections to be stored in a database
    # linked to all necessary ids only once.
    async def prepare_feeds(self) -> None:
        logger.info("preparing_feeds")

        feeds = []
        feeds_collections = []
        feed_infos = []

        source_uids = {
            feed_info.url: url_to_source_uid(feed_info.url)
            for collection in self._collections
            for feed_info in collection.feeds
        }

        source_ids = await get_source_ids(source_uids.values())

        # ensure that all feeds are really in the DB
        for collection in self._collections:
            for feed_info in collection.feeds:
                real_feed = Feed(
                    id=new_feed_id(),
                    source_id=source_ids[source_uids[feed_info.url]],
                    url=feed_info.url,
                    title=feed_info.title,
                    description=feed_info.description,
                )

                feed_infos.append(feed_info)
                feeds.append(real_feed)
                feeds_collections.append(collection.id)

        feed_ids = await f_domain.save_feeds(feeds)

        for feed_id, collection_id, feed_info in zip(feed_ids, feeds_collections, feed_infos):
            feed_info.feed_id = feed_id

            if feed_id not in self._feeds_in_collections:
                self._feeds_in_collections[feed_id] = set()

            self._feeds_in_collections[feed_id].add(collection_id)

        logger.info("feeds_prepared")

    def collections_for_feed(self, feed_id: FeedId) -> list[CollectionId]:
        return list(self._feeds_in_collections.get(feed_id, []))

    async def initialize(self, collection_configs: pathlib.Path | None = settings.collection_configs) -> None:
        logger.info("initializing_collections")

        if collection_configs is not None:
            self.load(collection_configs)

        self.validate_collection_ids()
        self.validate_collection_slugs()
        self.validate_collection_gui_order()
        self.validate_collection_is_not_empty()

        await self.prepare_feeds()

        self.initialized = True

        logger.info("collections_initialized")

    def has_feed(self, feed_id: FeedId) -> bool:
        return feed_id in self._feeds_in_collections

    # TODO: refactor collections logic to have only single feed object
    def get_feed_info(self, feed_id: FeedId) -> FeedInfo:
        for collection_id in self._feeds_in_collections.get(feed_id, ()):
            collection = self.collection(collection_id)

            for feed_info in collection.feeds:
                if feed_info.feed_id == feed_id:
                    return feed_info

        raise errors.FeedNotFound()

    async def add_test_collection(self, collection: Collection) -> None:
        self._collections.append(collection)
        await self.initialize(collection_configs=None)

    async def add_test_feed_to_collections(self, collection_id: CollectionId, feed_id: FeedId) -> None:

        feed = await f_domain.get_feed(feed_id)

        assert feed.title
        assert feed.description

        feed_info = FeedInfo(url=feed.url, title=feed.title, description=feed.description)

        for collection in self._collections:
            if collection.id == collection_id:
                collection.feeds.append(feed_info)
                break
        else:
            raise NotImplementedError("Collection not found")

        await self.initialize(collection_configs=None)


collections = Collections()



import asyncio
import math
from typing import Sequence

from ffun.core import logging
from ffun.library import domain as l_domain
from ffun.library.entities import Entry
from ffun.llms_framework import errors
from ffun.llms_framework.entities import (
    APIKeyUsage,
    ChatRequest,
    ChatResponse,
    LLMCollectionApiKey,
    LLMConfiguration,
    LLMGeneralApiKey,
    LLMTokens,
    SelectKeyContext,
    USDCost,
)
from ffun.llms_framework.keys_rotator import choose_api_key, use_api_key
from ffun.llms_framework.provider_interface import ProviderInterface

logger = logging.get_module_logger()


def split_text(text: str, parts: int, intersection: int) -> list[str]:
    if parts < 1:
        raise errors.TextPartsMustBePositive()

    if intersection < 0:
        raise errors.TextIntersectionMustBePositiveOrZero()

    if not text:
        raise errors.TextIsEmpty()

    if parts == 1:
        return [text]

    base_part_size = int(math.ceil(len(text) / parts))

    if base_part_size * parts - len(text) >= base_part_size:
        raise errors.TextIsTooShort()

    text_parts: list[str] = []

    index = 0

    while index < len(text):
        left_border = max(0, index - intersection)
        right_border = min(len(text), index + base_part_size + intersection)

        part = text[left_border:right_border]

        text_parts.append(part)

        index += base_part_size

    return text_parts


def split_text_according_to_tokens(  # noqa: CCR001
    llm: ProviderInterface, llm_config: LLMConfiguration, text: str
) -> list[str]:
    parts_number = 0

    model = llm.get_model(llm_config)

    while True:
        parts_number += 1

        parts = split_text(text, parts=parts_number, intersection=llm_config.text_parts_intersection)

        parts_tokens = [llm.estimate_tokens(llm_config, text=part) for part in parts]

        if any(tokens + llm_config.max_return_tokens >= model.max_context_size for tokens in parts_tokens):
            continue

        result: list[str] = []
        total_tokens = 0

        for part, tokens in zip(parts, parts_tokens):
            total_tokens += tokens + llm_config.max_return_tokens

            if total_tokens > model.max_tokens_per_entry:
                # trim the last part if the text exceeds the model limit
                logger.warning(
                    "entry_trimmed_for_llm",
                    model_provider=model.provider,
                    model_name=model.name,
                    model_max_tokens_per_entry=model.max_tokens_per_entry,
                )
                return result

            result.append(part)

        return result


async def search_for_api_key(
    llm: ProviderInterface,
    llm_config: LLMConfiguration,
    entry: Entry,
    requests: Sequence[ChatRequest],
    collections_api_key: LLMCollectionApiKey | None,
    general_api_key: LLMGeneralApiKey | None,
) -> APIKeyUsage | None:
    # TODO: here may be problems with too big context window for gemini
    #       (we'll reserve too much tokens), see ModelInfo.max_tokens_per_entry as a potential solution

    model = llm.get_model(llm_config)

    # TODO: test new calculation logic
    reserved_cost = USDCost(
        len(requests) * model.tokens_cost(input_tokens=model.max_context_size, output_tokens=model.max_return_tokens)
    )

    feed_ids = await l_domain.get_feeds_for_entry(entry.id)

    select_key_context = SelectKeyContext(
        llm_config=llm_config,
        feed_ids=feed_ids,
        entry_age=entry.age,
        reserved_cost=reserved_cost,
        collections_api_key=collections_api_key,
        general_api_key=general_api_key,
    )

    return await choose_api_key(llm, select_key_context)


# TODO: test tokens costs
async def call_llm(
    llm: ProviderInterface, llm_config: LLMConfiguration, api_key_usage: APIKeyUsage, requests: Sequence[ChatRequest]
) -> list[ChatResponse]:

    model = llm.get_model(llm_config)

    async with use_api_key(api_key_usage):
        tasks = [llm.chat_request(llm_config, api_key_usage.api_key, request) for request in requests]

        responses = await asyncio.gather(*tasks)

        api_key_usage.input_tokens = LLMTokens(sum(response.input_tokens() for response in responses))
        api_key_usage.output_tokens = LLMTokens(sum(response.output_tokens() for response in responses))

        api_key_usage.used_cost = model.tokens_cost(
            input_tokens=api_key_usage.input_tokens, output_tokens=api_key_usage.output_tokens
        )

    return responses



import datetime

from pytest_mock import MockerFixture

from ffun.llms_framework.entities import KeyStatus
from ffun.llms_framework.keys_statuses import Statuses


class TestStatuses:
    def test_inititalization(self, api_key_statuses: Statuses) -> None:
        assert api_key_statuses._statuses == {}

    def test_set_status(self, api_key_statuses: Statuses) -> None:
        status_1, status_2, *_ = list(KeyStatus)

        api_key_statuses.set("key_1", status_1)
        api_key_statuses.set("key_2", status_2)

        assert api_key_statuses.get("key_1") == status_1
        assert api_key_statuses.get("key_2") == status_2

    def test_get__no_key(self, api_key_statuses: Statuses) -> None:
        assert api_key_statuses.get("key_1") == KeyStatus.unknown

    def test_get__works(self, api_key_statuses: Statuses) -> None:
        api_key_statuses.set("key_1", KeyStatus.works)
        assert api_key_statuses.get("key_1") == KeyStatus.works

    def test_get__broken(self, api_key_statuses: Statuses) -> None:
        api_key_statuses.set("key_1", KeyStatus.broken)
        assert api_key_statuses.get("key_1") == KeyStatus.broken

    def test_get__broken_expired(self, api_key_statuses: Statuses, mocker: MockerFixture) -> None:
        api_key_statuses.set("key_1", KeyStatus.broken)

        mocker.patch("ffun.llms_framework.settings.settings.key_broken_timeout", datetime.timedelta(seconds=0))

        assert api_key_statuses.get("key_1") == KeyStatus.unknown

    def test_get__quota(self, api_key_statuses: Statuses) -> None:
        api_key_statuses.set("key_1", KeyStatus.quota)
        assert api_key_statuses.get("key_1") == KeyStatus.quota

    def test_get__quota_expired(self, api_key_statuses: Statuses, mocker: MockerFixture) -> None:
        api_key_statuses.set("key_1", KeyStatus.quota)

        mocker.patch("ffun.llms_framework.settings.settings.key_quota_timeout", datetime.timedelta(seconds=0))

        assert api_key_statuses.get("key_1") == KeyStatus.unknown
